#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

from collections import deque
import math

from chorus import topology
import chorus.util.geometry as gm

# BOND_LENGTH = 1


def calc2dcoords(mol):
    """ Calculate optimal 2D coordinates of chemical structure
    """
    topology.recognize(mol)
    g = set(i for i, _ in mol.atoms_iter())
    # 1: get nodes in scaffolds
    scaffolds = []
    belongs = {}
    for i, rkeys in enumerate(sorted(mol.scaffolds, key=len)):
        scf = []
        for rkey in rkeys:
            ring = mol.rings[rkey]
            for r in ring:
                belongs[r] = i
            scf.append(ring)
            g -= set(ring)
        scaffolds.append(scf)
    # 2: traverse nodes and scaffolds
    # the node and scaffold graph should be a tree (no cycles)
    f = True
    # print(scaffolds)
    coords = {}
    while g:
        if f and scaffolds:  # largest scaffold is first
            stack = [scaffolds[-1][0][0]]
            f = False
        else:
            stack = [g.pop()]
        pred = {}
        branch = {}
        while stack:
            # print("stack: {}".format(stack))
            tail = stack.pop()
            # print("tail: {}".format(tail))
            if tail in belongs:  # scaffolds
                scf = scaffold_coords(scaffolds[belongs[tail]])
                # print(scf.keys())
                # rotate and translate
                if not coords:
                    coords = scf
                else:
                    u = coords[pred[tail]]
                    v = scf[tail]
                    op = [u[0] + math.cos(u[2]), u[1] + math.sin(u[2])]
                    translate(scf, gm.vector(v[:2], op))
                    rotate(scf, op, gm.rad(u[2] + math.pi - v[2]))
                    coords.update(scf)
                # stack nbrs of scaffold
                for k in scf.keys():
                    pred[k] = None
                    for nbr in mol.neighbors(k):
                        if nbr not in scf.keys():
                            stack.append(nbr)
                            pred[nbr] = k
            else:  # append linker
                if tail not in pred:  # isolated
                    coords[tail] = [0, 0, 0, 1]
                    continue
                p = pred[tail]
                x, y, ang, d = coords[p]
                # TODO: ring configuration
                coords[tail] = [x + math.cos(ang), y + math.sin(ang),
                                ang + d * math.pi / 3, d * -1]
                if p not in branch:
                    coords[p][2] = gm.rad(coords[p][2] + math.pi * 2 / 3 * d)
                    branch[p] = 1
                elif branch[p] == 1:
                    coords[p][2] = gm.rad(coords[p][2] + math.pi * d)
                    branch[p] += 1
                elif branch[p] == 2:
                    coords[p][2] = gm.rad(coords[p][2] + math.pi * 2 / 3 * d)
                    branch[p] += 1
                for nbr in mol.neighbors(tail):
                    if nbr not in pred:
                        stack.append(nbr)
                        pred[nbr] = tail
        g -= set(pred)
    resolve_overlap(coords)
    for i, a in mol.atoms_iter():
        mol.atom(i).coords = coords[i][:2]


def translate(vertices, operator):
    for k, v in vertices.items():
        new_pos = list(gm.translate(v[:2], operator))
        vertices[k] = new_pos + v[2:]


def rotate(vertices, center, angle):
    for k, v in vertices.items():
        new_pos = list(gm.rotate(v[:2], angle, center))
        new_angle = v[2] + angle
        vertices[k] = [new_pos[0], new_pos[1], new_angle, v[3]]


def scaffold_coords(rings):
    """ assign scaffold coordinate and angle
    {node: (coords, angle)}
    """
    rs = deque(sorted(rings, key=len, reverse=True))
    base = rs.popleft()
    first = base[0]
    coords = {first: [0, 0, 0, 1]}
    attach_spiro(coords, base, first)
    # display(coords)
    while rs:
        r = rs.popleft()
        isec = list(set(coords) & set(r))
        if not isec:
            rs.append(r)
        elif len(isec) == 1:  # spiro
            attach_spiro(coords, r, isec)
        elif len(isec) == 2:  # fuse
            attach_fused(coords, r, isec)
        else:  # bridge or append
            bridge(coords, r, isec)
        # display(coords)
    return coords


def bridge(coords, ring, isec):
    r = deque(ring)
    while r[0] not in isec or r[-1] in isec:
        r.rotate()
    v = r.popleft()
    u = None
    while True:
        p = r.popleft()
        if p in isec:
            u = p
        else:
            break
    r.appendleft(p)  # TODO: refactor
    ux, uy, ua, _ = coords[u]
    vx, vy, va, _ = coords[v]
    utov = gm.vector((ux, uy), (vx, vy))
    utovang = gm.rad(math.atan2(utov[1], utov[0]))
    dist = gm.distance((ux, uy), (vx, vy))
    step, angle = bridge_angle(dist, len(r))
    angle += utovang
    prev = u
    coords[prev][2] = gm.rad(utovang + math.pi)  # reverse
    while r:
        tail = r.popleft()
        x, y, a, _ = coords[prev]
        coords[tail] = [x + math.cos(angle), y + math.sin(angle),
                        gm.rad(a + step), 1]
        angle += step
        prev = tail


def bridge_angle(base, n):
    # TODO: There may be an elegant analytical solution
    length = 1
    res = 0
    dmin = 100
    for i in range(330):
        r = math.pi / 180 * i
        a = (math.pi - r / 2) / (n + 1)
        d = abs(length * math.sin(r / 2) - base * math.sin(a))
        if d < dmin:
            res = r
            dmin = d
        if d < 0.0001:
            break
    step = -(2 * math.pi - res) / (n + 1)
    angle = -step * n / 2
    return step, angle


def attach_spiro(coords, ring, isec):
    r = deque(ring)
    while r[0] != isec:
        r.rotate()
    prev = r.popleft()
    step = math.pi * 2 / len(ring)  # angle shift
    angle = gm.rad(coords[prev][2] + (math.pi - step) / 2)  # init angle
    coords[prev][2] += math.pi  # reverse
    while r:
        tail = r.popleft()
        x, y, a, _ = coords[prev]
        coords[tail] = [x + math.cos(angle), y + math.sin(angle),
                        gm.rad(a - step), 1]
        angle -= step
        prev = tail


def attach_fused(coords, ring, isec):
    r = deque(ring)
    while r[0] not in isec or r[-1] in isec:
        r.rotate()
    first = r.popleft()
    second = r.popleft()
    x, y = gm.vector(coords[first][:2], coords[second][:2])
    ftos = gm.rad(math.atan2(y, x))
    if ftos >= coords[second][2] or coords[second][2] > ftos + math.pi:
        # clockwise
        step = math.pi * -2 / len(ring)
        coords[second][2] = ftos + (math.pi + step) / 2
    else:  # counter-clockwise
        step = math.pi * 2 / len(ring)
        coords[second][2] = ftos - (math.pi - step) / 2
    angle = ftos + step
    prev = second
    while r:
        tail = r.popleft()
        x, y, a, _ = coords[prev]
        coords[tail] = [x + math.cos(angle), y + math.sin(angle),
                        gm.rad(a + step), 1]
        angle += step
        prev = tail
    coords[second][2] = ftos
    coords[first][2] = gm.rad(ftos + math.pi)


def display(coords):
    print("")
    for k, v in coords.items():
        print("{}: ({}, {}) {} {}".format(k,
                                          round(v[0], 2),
                                          round(v[1], 2),
                                          round(v[2], 2), v[3]))


def resolve_overlap(coords):
    while True:
        break



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

import chorus.util.geometry as gm

# Multiple bond interval / length
F_MULI = 0.15
# 1 - (inner / outer) double bond length
F_INNL = 0.2
# Factor to avoid overlap of bond segment with atom symbol
F_AOVL = 0.3


def draw(canvas, mol):
    """Draw molecule structure image.

    Args:
        canvas: draw.drawable.Drawable
        mol: model.graphmol.Compound
    """
    mol.require("ScaleAndCenter")
    mlb = mol.size2d[2]
    if not mol.atom_count():
        return
    bond_type_fn = {
        1: {
            0: single_bond,
            1: wedged_single,
            2: dashed_wedged_single,
            3: wave_single,
        }, 2: {
            0: cw_double,
            1: counter_cw_double,
            2: double_bond,
            3: cross_double
        }, 3: {
            0: triple_bond
        }
    }
    # Draw bonds
    for u, v, bond in mol.bonds_iter():
        if not bond.visible:
            continue
        if (u < v) == bond.is_lower_first:
            f, s = (u, v)
        else:
            s, f = (u, v)
        p1 = mol.atom(f).coords
        p2 = mol.atom(s).coords
        if p1 == p2:
            continue  # avoid zero division
        if mol.atom(f).visible:
            p1 = gm.t_seg(p1, p2, F_AOVL, 2)[0]
        if mol.atom(s).visible:
            p2 = gm.t_seg(p1, p2, F_AOVL, 1)[1]
        color1 = mol.atom(f).color
        color2 = mol.atom(s).color
        bond_type_fn[bond.order][bond.type](
            canvas, p1, p2, color1, color2, mlb)

    # Draw atoms
    for n, atom in mol.atoms_iter():
        if not atom.visible:
            continue
        p = atom.coords
        color = atom.color
        # Determine text direction
        if atom.H_count:
            cosnbrs = []
            hrzn = (p[0] + 1, p[1])
            for nbr in mol.graph.neighbors(n):
                pnbr = mol.atom(nbr).coords
                try:
                    cosnbrs.append(gm.dot_product(hrzn, pnbr, p) /
                                   gm.distance(p, pnbr))
                except ZeroDivisionError:
                    pass
            if not cosnbrs or min(cosnbrs) > 0:
                # [atom]< or isolated node(ex. H2O, HCl)
                text = atom.formula_html(True)
                canvas.draw_text(p, text, color, "right")
                continue
            elif max(cosnbrs) < 0:
                # >[atom]
                text = atom.formula_html()
                canvas.draw_text(p, text, color, "left")
                continue
        # -[atom]- or no hydrogens
        text = atom.formula_html()
        canvas.draw_text(p, text, color, "center")


def single_bond(canvas, p1, p2, color1, color2, mlb):
    canvas.draw_line(p1, p2, color1, color2)


def wedged_single(canvas, p1, p2, color1, color2, mlb):
    canvas.draw_wedge(p1, p2, color1)


def dashed_wedged_single(canvas, p1, p2, color1, color2, mlb):
    canvas.draw_dashed_wedge(p1, p2, color1)


def wave_single(canvas, p1, p2, color1, color2, mlb):
    canvas.draw_wave_line(p1, p2, color1)


def double_bond(canvas, p1, p2, color1, color2, mlb):
    ip1, ip2 = gm.p_seg(p1, p2, True, F_MULI / 2 * mlb)
    op1, op2 = gm.p_seg(p1, p2, False, F_MULI / 2 * mlb)
    canvas.draw_line(ip1, ip2, color1, color2)
    canvas.draw_line(op1, op2, color1, color2)


def cw_double(canvas, p1, p2, color1, color2, mlb):
    ip1, ip2 = gm.p_seg(p1, p2, True, F_MULI * mlb, F_INNL)
    canvas.draw_line(p1, p2, color1, color2)
    canvas.draw_line(ip1, ip2, color1, color2)


def counter_cw_double(canvas, p1, p2, color1, color2, mlb):
    ip1, ip2 = gm.p_seg(p1, p2, False, F_MULI * mlb, F_INNL)
    canvas.draw_line(p1, p2, color1, color2)
    canvas.draw_line(ip1, ip2, color1, color2)


def cross_double(canvas, p1, p2, color1, color2, mlb):
    ip1, ip2 = gm.p_seg(p1, p2, True, F_MULI / 2 * mlb)
    op1, op2 = gm.p_seg(p1, p2, False, F_MULI / 2 * mlb)
    canvas.draw_line(ip1, op2, color1, color2)
    canvas.draw_line(op1, ip2, color1, color2)


def triple_bond(canvas, p1, p2, color1, color2, mlb):
    ip1, ip2 = gm.p_seg(p1, p2, True, F_MULI * mlb)
    op1, op2 = gm.p_seg(p1, p2, False, F_MULI * mlb)
    canvas.draw_line(p1, p2, color1, color2)
    canvas.draw_line(ip1, ip2, color1, color2)
    canvas.draw_line(op1, op2, color1, color2)



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

import traceback
import re

from chorus.model.atom import Atom
from chorus.model.bond import Bond
from chorus.model.graphmol import Compound
from chorus import molutil
import chorus.util.text as tx


def inspect(lines):
    """Inspect SDFile list of string

    Returns:
        tuple: (data label list, number of records)
    """
    labels = set()
    count = 0
    exp = re.compile(r">.*?<([\w ]+)>")  # Space should be accepted
    valid = False
    for line in lines:
        if line.startswith("M  END\n"):
            valid = True
        elif line.startswith("$$$$"):
            count += 1
            valid = False
        else:
            result = exp.match(line)
            if result:
                labels.add(result.group(1))
    if valid:
        count += 1
    return list(labels), count


def inspect_text(text):
    # Lazy line splitter. More efficient memory usage than str.split.
    exp = re.compile(r"[^\n]*\n|.")
    lines = (x.group(0) for x in re.finditer(exp, text))
    return inspect(lines)


def inspect_file(path):
    """Inspect SDFile structure

    Returns:
        tuple: (data label list, number of records)
    """
    with open(path, 'rb') as f:
        labels, count = inspect(tx.decode(line) for line in f)
    return labels, count


def optional_data(lines):
    """Parse SDFile data part into dict"""
    data = {}
    exp = re.compile(r">.*?<([\w ]+)>")  # Space should be accepted
    for i, line in enumerate(lines):
        result = exp.match(line)
        if result:
            data[result.group(1)] = lines[i + 1]
    return data


def atoms(lines):
    """Parse atom block into atom objects

    Returns:
        dict: networkx nodes
    """
    # Convert sdf style charge to actual charge
    conv_charge_table = {0: 0, 1: 3, 2: 2, 3: 1, 4: 0, 5: -1, 6: -2, 7: -3}
    results = {}
    for i, line in enumerate(lines):
        symbol = line[31:34].rstrip()
        try:
            atom = Atom(symbol)
        except KeyError:
            raise ValueError(symbol)
        xpos = float(line[0:10])
        ypos = float(line[10:20])
        zpos = float(line[20:30])
        atom.coords = (xpos, ypos, zpos)
        atom.mass_diff = int(line[34:37])
        old_sdf_charge = int(line[37:40])
        atom.charge = conv_charge_table[old_sdf_charge]
        if old_sdf_charge == 4:
            atom.radical = 1
        # atom.stereo_flag = int(line[40:43])  # Not used
        # valence = int(line[46:49])
        # if valence:
        #     atom.valence = valence
        results[i + 1] = {"atom": atom}
    return results


def bonds(lines, atoms):
    """Parse bond block into bond objects

    Returns:
        dict: networkx adjacency dict
    """
    # Convert sdf style stereobond (see chem.model.bond.Bond)
    conv_stereo_table = {0: 0, 1: 1, 3: 3, 4: 3, 6: 2}
    results = {a: {} for a in atoms}
    for line in lines:
        bond = Bond()
        first = int(line[0:3])
        second = int(line[3:6])
        if first > second:
            bond.is_lower_first = 0
        order = int(line[6:9])
        if order < 4:
            bond.order = order
        bond.type = conv_stereo_table[int(line[9:12])]
        results[first][second] = {"bond": bond}
        results[second][first] = {"bond": bond}
    return results


def properties(lines):
    """Parse properties block

    Returns:
        dict: {property_type: (atom_index, value)}
    """
    results = {}
    for i, line in enumerate(lines):
        type_ = line[3:6]
        if type_ not in ["CHG", "RAD", "ISO"]:
            continue  # Other properties are not supported yet
        count = int(line[6:9])
        results[type_] = []
        for j in range(count):
            idx = int(line[10 + j * 8: 13 + j * 8])
            val = int(line[14 + j * 8: 17 + j * 8])
            results[type_].append((idx, val))
    return results


def add_properties(props, mol):
    """apply properties to the molecule object

    Returns:
        None (alter molecule object directly)
    """
    if not props:
        return
    # The properties supersedes all charge and radical values in the atom block
    for _, atom in mol.atoms_iter():
        atom.charge = 0
        atom.multi = 1
        atom.mass = None
    for prop in props.get("CHG", []):
        mol.atom(prop[0]).charge = prop[1]
    for prop in props.get("RAD", []):
        mol.atom(prop[0]).multi = prop[1]
    for prop in props.get("ISO", []):
        mol.atom(prop[0]).mass = prop[1]


def molecule(lines):
    """Parse molfile part into molecule object

    Args:
        lines (list): lines of molfile part

    Raises:
        ValueError: Symbol not defined in periodictable.yaml
                    (Polymer expression not supported yet)
    """
    count_line = lines[3]
    num_atoms = int(count_line[0:3])
    num_bonds = int(count_line[3:6])
    # chiral_flag = int(count_line[12:15])  # Not used
    # num_prop = int(count_line[30:33])  # "No longer supported"
    compound = Compound()
    compound.graph._node = atoms(lines[4: num_atoms+4])
    compound.graph._adj = bonds(lines[num_atoms+4: num_atoms+num_bonds+4],
                                compound.graph._node.keys())
    props = properties(lines[num_atoms+num_bonds+4:])
    add_properties(props, compound)
    return compound


def mol_supplier(lines, no_halt, assign_descriptors):
    """Yields molecules generated from CTAB text

    Args:
        lines (iterable): CTAB text lines
        no_halt (boolean):
            True: shows warning messages for invalid format and go on.
            False: throws an exception for it and stop parsing.
        assign_descriptors (boolean):
            if True, default descriptors are automatically assigned.
    """
    def sdf_block(lns):
        mol = []
        opt = []
        is_mol = True
        for line in lns:
            if line.startswith("$$$$"):
                yield mol[:], opt[:]
                is_mol = True
                mol.clear()
                opt.clear()
            elif line.startswith("M  END"):
                is_mol = False
            elif is_mol:
                mol.append(line.rstrip())
            else:
                opt.append(line.rstrip())
        if mol:
            yield mol, opt

    for i, (mol, opt) in enumerate(sdf_block(lines)):
        try:
            c = molecule(mol)
            if assign_descriptors:
                molutil.assign_descriptors(c)
        except ValueError as err:
            if no_halt:
                print("Unsupported symbol: {} (#{} in v2000reader)".format(
                      err, i + 1))
                c = molutil.null_molecule(assign_descriptors)
            else:
                raise ValueError("Unsupported symbol: {}".format(err))
        except:
            if no_halt:
                print("Unexpected error (#{} in v2000reader)".format(i + 1))
                c = molutil.null_molecule(assign_descriptors)
                c.data = optional_data(opt)
                yield c
                continue
            else:
                print(traceback.format_exc())
                raise Exception("Unsupported Error")
        c.data = optional_data(opt)
        yield c


def mols_from_text(text, no_halt=True, assign_descriptors=True):
    """Returns molecules generated from sdfile text

    Throws:
        StopIteration: if the text does not have molecule
        ValueError: if Unsupported symbol is found
    """
    if isinstance(text, bytes):
        t = tx.decode(text)
    else:
        t = text
    # Lazy line splitter. More efficient memory usage than str.split.
    exp = re.compile(r"[^\n]*\n|.")
    sp = (x.group(0) for x in re.finditer(exp, t))
    for c in mol_supplier(sp, no_halt, assign_descriptors):
        yield c


def mol_from_text(text, assign_descriptors=True):
    """Parse CTAB text and return first one as a Compound object.

    Throws:
        StopIteration: if the text does not have molecule
        ValueError: if Unsupported symbol is found
    """
    cg = mols_from_text(text, False, assign_descriptors)
    return next(cg)


def mols_from_file(path, no_halt=True, assign_descriptors=True):
    """Compound supplier from CTAB text file (.mol, .sdf)"""
    with open(path, 'rb') as f:
        fd = (tx.decode(line) for line in f)
        for c in mol_supplier(fd, no_halt, assign_descriptors):
            yield c


def mol_from_file(path, assign_descriptors=True):
    """Parse CTAB file and return first one as a Compound object."""
    cs = mols_from_file(path, False, assign_descriptors)
    return next(cs)



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

import itertools
import math
import statistics

from chorus.util import geometry
from chorus.util import iterator
from chorus import molutil


def display_terminal_carbon(mol):
    """Set visible=True to the terminal carbon atoms.
    """
    for i, a in mol.atoms_iter():
        if mol.neighbor_count(i) == 1:
            a.visible = True


def equalize_terminal_double_bond(mol):
    """Show equalized double bond if it is connected to terminal atom.
    """
    for i, a in mol.atoms_iter():
        if mol.neighbor_count(i) == 1:
            nb = list(mol.neighbors(i).values())[0]
            if nb.order == 2:
                nb.type = 2


def spine_to_terminal_wedge(mol):
    """Arrange stereo wedge direction from spine to terminal atom
    """
    for i, a in mol.atoms_iter():
        if mol.neighbor_count(i) == 1:
            ni, nb = list(mol.neighbors(i).items())[0]
            if nb.order == 1 and nb.type in (1, 2) \
                    and ni > i != nb.is_lower_first:
                nb.is_lower_first = not nb.is_lower_first
                nb.type = {1: 2, 2: 1}[nb.type]


def format_ring_double_bond(mol):
    """Set double bonds around the ring.
    """
    mol.require("Topology")
    mol.require("ScaleAndCenter")
    for r in sorted(mol.rings, key=len, reverse=True):
        vertices = [mol.atom(n).coords for n in r]
        try:
            if geometry.is_clockwise(vertices):
                cpath = iterator.consecutive(itertools.cycle(r), 2)
            else:
                cpath = iterator.consecutive(itertools.cycle(reversed(r)), 2)
        except ValueError:
            continue
        for _ in r:
            u, v = next(cpath)
            b = mol.bond(u, v)
            if b.order == 2:
                b.type = int((u > v) == b.is_lower_first)


def scale_and_center(mol):
    """Center and Scale molecule 2D coordinates.
    This method changes mol coordinates directly to center but not scale.
    This method returns width, height and MLB(median length of bond)
    and scaling will be done by drawer method with these values.

    Returns:
        width: float
        height: float
        mlb: median length of bond
    """
    cnt = mol.atom_count()
    if cnt < 2:
        mol.size2d = (0, 0, 1)
        mol.descriptors.add("ScaleAndCenter")
        return
    xs = []
    ys = []
    for _, atom in mol.atoms_iter():
        xs.append(atom.coords[0])
        ys.append(atom.coords[1])
    xmin, xmax = (min(xs), max(xs))
    ymin, ymax = (min(ys), max(ys))
    width = xmax - xmin
    height = ymax - ymin
    x_offset = width / 2 + xmin
    y_offset = height / 2 + ymin
    dists = []
    for u, v, _ in mol.bonds_iter():
        dists.append(geometry.distance(mol.atom(u).coords, mol.atom(v).coords))
    if not len(dists):  # No connection
        mlb = math.sqrt(max([width, height]) / cnt)  # empirical
    elif not max(dists):  # All connected atoms are overlapped
        mol.size2d = (0, 0, 1)
        mol.descriptors.add("ScaleAndCenter")
        return
    else:
        mlb = statistics.median(dists)
    # Centering
    for _, atom in mol.atoms_iter():
        atom.coords = (atom.coords[0] - x_offset, atom.coords[1] - y_offset)
    mol.size2d = (width, height, mlb)
    mol.descriptors.add("ScaleAndCenter")


def ready_to_draw(mol):
    """Shortcut function to prepare molecule to draw.
    Overwrite this function for customized appearance.
    It is recommended to clone the molecule before draw
    because all the methods above are destructive.
    """
    copied = molutil.clone(mol)
    # display_terminal_carbon(mol)
    equalize_terminal_double_bond(copied)
    # spine_to_terminal_wedge(copied)
    scale_and_center(copied)
    format_ring_double_bond(copied)
    return copied



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

from math import cos, acos, sin, hypot, pi
from itertools import cycle

from chorus.util import iterator


def vector(p1, p2):
    """ p1 to p2 vector
    Args:
      p1, p2: point (x, y)
    """
    return p2[0] - p1[0], p2[1] - p1[1]


def distance(p1, p2):
    """ distance between p1 and p2 """
    return hypot(*vector(p1, p2))


def translate(v, op):
    """ translate vector
    Args:
      v: vector (x, y)
      op: operator (x, y)
    """
    return v[0] + op[0], v[1] + op[1]


def scale(p, factor, o=(0, 0)):
    """ scale vector
    Args:
      p: point (x, y)
      factor: scaling factor
      o: origin (x, y)
    """
    v = vector(o, p)
    sv = v[0] * factor, v[1] * factor
    return translate(sv, o)


def unit(v, lg=1):
    """ unit vector
    Args:
        v: vector (x, y)
        lg: length
    Raises:
        ValueError: Null vector was given
    """
    try:
        res = scale(v, lg / distance((0, 0), v))
    except ZeroDivisionError:
        raise ValueError("Null vector was given")
    return res


def rotate(p, rad, o=(0, 0)):
    """ rotate vector
    Args:
      p: point (x, y)
      rad: angle(radian)
      o: origin (x, y)
    """
    v = vector(o, p)
    fx = lambda x, y, d: x * cos(d) - y * sin(d)
    fy = lambda x, y, d: x * sin(d) + y * cos(d)
    rv = fx(v[0], v[1], rad), fy(v[0], v[1], rad)
    return translate(rv, o)


def cross_product(p1, p2, o=(0, 0)):
    """ Returns cross product
    Args:
      p1, p2: point (x, y)
      o: origin
    """
    v1 = vector(o, p1)
    v2 = vector(o, p2)
    return v1[0] * v2[1] - v1[1] * v2[0]


def dot_product(p1, p2, o=(0, 0)):
    """ Returns dot product
    Args:
      p1, p2: point (x, y)
      o: origin
    """
    v1 = vector(o, p1)
    v2 = vector(o, p2)
    return v1[0] * v2[0] + v1[1] * v2[1]


def interior_angle(p1, p2, o=(0, 0)):
    """ Returns interior angle of two vector(0 <= θ <= pi)
    Args:
      p1, p2: point (x, y)
      o: origin
    Raises:
      ValueError: p1 or p2 is overlapped with origin
    """
    v1 = vector(o, p1)
    v2 = vector(o, p2)
    len1 = distance(o, p1)
    len2 = distance(o, p2)
    try:
        return acos(dot_product(v1, v2) / (len1 * len2))
    except ZeroDivisionError:
        raise ValueError("p1 or p2 is overlapped with origin")


def m_seg(p1, p2, rad, dist):
    """ move segment by distance
    Args:
      p1, p2: point(x, y)
      rad: relative direction angle(radian)
      dist: distance
    Return:
      translated segment(p1, p2)
    """
    v = vector(p1, p2)
    m = unit(rotate(v, rad), dist)
    return translate(p1, m), translate(p2, m)


def t_seg(p1, p2, t, align=0):
    """ trim segment
    Args:
      p1, p2: point(x, y)
      t: scaling factor (1 - trimed segment / original segment)
      align: 1: trim p2, 2: trim p1, 0: both side
    Return:
      trimmed segment(p1, p2)
    """
    v = vector(p1, p2)
    result = {
        1: lambda a, b: (a, translate(b, scale(v, -t))),
        2: lambda a, b: (translate(a, scale(v, t)), b),
        0: lambda a, b: (translate(a, scale(v, t / 2)),
                         translate(b, scale(v, -t / 2)))
    }
    return result[align](p1, p2)


def p_seg(p1, p2, cw, interval, trim=0, align=0):
    """ parallel segment
    Args:
      p1, p2: point(x, y)
      cw: m_seg rad True: -π/2, False: π/2
      interval: m_seg dist
      trim: t_seg trim
      align: t_seg align
    """
    case = {True: pi / -2, False: pi / 2}
    p1m, p2m = m_seg(p1, p2, case[cw], interval)
    return t_seg(p1m, p2m, trim, align)


def is_clockwise(vertices):
    """ Evaluate whether vertices are in clockwise order.
    Args:
      vertices: list of vertices (x, y) in polygon.
    Returns:
      True: clockwise, False: counter-clockwise
    Raises:
      ValueError: the polygon is complex or overlapped.
    """
    it = iterator.consecutive(cycle(vertices), 3)
    clockwise = 0
    counter = 0
    for _ in range(len(vertices)):
        p0, p1, p2 = next(it)
        cross = cross_product(p1, p2, p0)
        int_angle = interior_angle(p0, p2, p1)  # raises ValueError
        if cross < 0:
            clockwise += int_angle
            counter += 2 * pi - int_angle
        else:
            clockwise += 2 * pi - int_angle
            counter += int_angle
    if round(clockwise / pi) == len(vertices) - 2:
        return True
    elif round(counter / pi) == len(vertices) - 2:
        return False
    else:
        raise ValueError("the polygon is complex or overlapped")


def rad(radian):
    """ A radian in the range of 0 - 2pai """
    return (radian + 2 * pi) % (2 * pi)



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

import glob
import os

RESOURCE_DIR = os.path.join(os.path.dirname(__file__), "resources")

MOL = {}
for path in glob.glob(os.path.join(RESOURCE_DIR, "test/*.mol")):
    name = os.path.basename(path).split(".")[0]
    with open(path) as file:
        MOL[name] = file.read()

for path in glob.glob(os.path.join(RESOURCE_DIR, "DrugBank/*.mol")):
    name = os.path.basename(path).split(".")[0]
    with open(path) as file:
        MOL[name] = file.read()

for path in glob.glob(os.path.join(RESOURCE_DIR, "PubChem/*.sdf")):
    name = os.path.basename(path).split(".")[0]
    with open(path) as file:
        MOL[name] = file.read()



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

import re


def substitute(dict_, source):
    """ Perform re.sub with the patterns in the given dict
    Args:
      dict_: {pattern: repl}
      source: str
    """
    d_esc = (re.escape(k) for k in dict_.keys())
    pattern = re.compile('|'.join(d_esc))
    return pattern.sub(lambda x: dict_[x.group()], source)


def decode(byte_):
    try:
        res = byte_.decode('utf-8')
    except UnicodeDecodeError:
        try:
            res = byte_.decode("cp1252")
        except UnicodeDecodeError:
            try:
                res = byte_.decode("shift-jis")
            except UnicodeDecodeError:
                raise ValueError("Unsupported codec")
    return res


def decode_file(path):
    with open(path, 'rb') as f:
        for line in f:
            yield decode(line)



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

import os
import sys
import cProfile
import pstats
import linecache
import tracemalloc
from itertools import chain
from collections import deque


def profile(func):
    """ Decorator
    Execute cProfile
    """
    def _f(*args, **kwargs):
        print("\n<<<---")
        pr = cProfile.Profile()
        pr.enable()
        res = func(*args, **kwargs)
        p = pstats.Stats(pr)
        p.strip_dirs().sort_stats('cumtime').print_stats(20)
        print("\n--->>>")
        return res
    return _f


def total_size(obj, verbose=False):
    """ Returns approximate memory size"""
    seen = set()

    def sizeof(o):
        if id(o) in seen:
            return 0
        seen.add(id(o))
        s = sys.getsizeof(o, default=0)
        if verbose:
            print(s, type(o), repr(o))
        if isinstance(o, (tuple, list, set, frozenset, deque)):
            s += sum(map(sizeof, iter(o)))
        elif isinstance(o, dict):
            s += sum(map(sizeof, chain.from_iterable(o.items())))
        elif "__dict__" in dir(o):
            s += sum(map(sizeof, chain.from_iterable(o.__dict__.items())))
        return s

    return sizeof(obj)


def malloc(func):
    """ Decorator
    Execute tracemalloc
    """
    def _f(*args, **kwargs):
        print("\n<<<---")
        tracemalloc.start()
        res = func(*args, **kwargs)
        snapshot = tracemalloc.take_snapshot()
        top_stats = snapshot.statistics('lineno')
        print("[ Top 10 ]")
        for i, stat in enumerate(top_stats[:10]):
            frame = stat.traceback[0]
            filename = os.sep.join(frame.filename.split(os.sep)[-2:])
            print("#%s: %s:%s: %.1f KiB"
                  % (i, filename, frame.lineno, stat.size / 1024))
            print(linecache.getline(frame.filename, frame.lineno).strip())
            # print(stat)
        print("--->>>\n")
        return res
    return _f


def malloc_diff(func):
    """ Decorator
    Execute tracemalloc
    """
    def _f(*args, **kwargs):
        print("\n<<<---")
        tracemalloc.start()
        snapshot1 = tracemalloc.take_snapshot()
        res = func(*args, **kwargs)
        snapshot2 = tracemalloc.take_snapshot()
        top_stats = snapshot2.compare_to(snapshot1, 'lineno')
        print("[ Top 10 differences]")
        for stat in top_stats[:10]:
            print(stat)
        print("--->>>\n")
        return res
    return _f


def mute(func):
    """ Decorator
    Make stdout silent
    """
    def _f(*args, **kwargs):
        sys.stdout = open(os.devnull, 'w')
        res = func(*args, **kwargs)
        sys.stdout.close()
        sys.stdout = sys.__stdout__
        return res
    return _f



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

from chorus.model.Abstractatom import AbstractAtom


def collapse(self):
    """ recognize sugar moiety and convert into Sugar object """
    # TODO: not implemented
    pass


class Sugar(AbstractAtom):
    def __init__(self):
        self.symbol = None
        self.composition
        self.mw

    def formula_html(self):
        return self.symbol

    def composition(self):
        return self.composition

    def mw(self):
        return self.mw


class Hexose(Sugar):
    def __init__(self):
        self.name = "Hexose"
        self.symbol = "Hex"
        self.composition = None
        self.mw = None



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

""" Maximal Common Substructure with Diameter Restriction (MCS-DR)

This calculetes maximum common edge subgraph (MCES) between two molecule
by using the algorithm based on maximum clique detection in the modular
product graph of the two molecule.
Diameter restriction is used for fast and stable local MCS estimation
and for common substructure mapping

"""

import networkx as nx
import numpy as np

from chorus.model.graphmol import Compound
from chorus import molutil
from chorus import remover

try:
    from chorus.cython import mcsdr
    find_cliques = mcsdr.find_cliques
    comparison_graph = mcsdr.comparison_graph
    CYTHON_AVAILABLE = True
except ImportError:
    from networkx.algorithms.clique import find_cliques
    CYTHON_AVAILABLE = False
    try:
        import numexpr as ne
        NUMEXPR_AVAILABLE = True
    except ImportError:
        NUMEXPR_AVAILABLE = False


def comparison_array(molecule, diameter=8, size=30,
                     ignore_hydrogen=True, timeout=1):
    """ Generate comparison array
    Comparison array consists of node pairs in the graph and a collater.
    42 bit collater
        6 bit of distance (0-31)
        18 bit of bond attribute x2
            9 bit of atom attribute x 2
                7 bit of atom number (0-127)
                2 bit of atom pi(0-3)


    [Sheridan, R.P. and Miller, M.D., J. Chem. Inf. Comput. Sci. 38 (1998) 915]

    Using possible_path_length instead of shortest_path_length
    and find intersection of the distance set each other,
    it is possible to detect roundabout matching path.
    Therefore, exact graph isomorphism can be determined.
    However, set intersection is too costful to do
    in spite of a trivial improvement.

    Args:
        mol: Compound
        cutoff: more distant connection than the value is no longer used
                for comparison matrix graph due to performance reason.

    Returns:
        arr(list): comparison array. list of tuple (node1, node2, collater)
        max_mcs(int): maximum size of possible mcs
        int_to_node(dict): int -> node index pair reconverter dict
    Throws:
        ValueError: if len(mol) < 3
    """
    molecule.require("Valence")
    mol = molutil.clone(molecule)
    if ignore_hydrogen:
        mol = molutil.make_Hs_implicit(mol)
    # Ignore salt, water
    remover.remove_salt(mol)
    remover.remove_water(mol)
    # multivalent coordinated metals notably affect the performance
    remover.remove_coordinated_metal(mol)
    g = nx.line_graph(mol.graph)
    node_to_int = {}
    for i, e in enumerate(g.nodes()):
        node_to_int[e] = i
        a1 = mol.atom(e[0])
        a2 = mol.atom(e[1])
        a1t = a1.number << 2 | a1.pi
        a2t = a2.number << 2 | a2.pi
        pair = sorted((a1t, a2t))
        g.node[e]["type"] = pair[0] << 9 | pair[1]
    # convert node index pair to integer expression
    g = nx.relabel_nodes(g, node_to_int)
    # interger -> index pair reconverter
    int_to_node = {v: k for k, v in node_to_int.items()}
    arr = []
    matrix = nx.Graph()
    for ui, ua in g.nodes(data=True):
        r = _reachables(g, ui, diameter, size)
        for vi, d in r.items():
            if not d:
                continue
            matrix.add_edge(ui, vi)
            code = (d << 18 | ua["type"]) << 18 | g.node[vi]["type"]
            arr.append((ui, vi, code))
    cliques, elapsed = find_cliques(matrix, timeout=timeout)
    if elapsed > timeout:
        raise RuntimeError("Max fragment determination has timed out")
    max_size = len(max(cliques, key=len, default=[]))
    return arr, max_size, int_to_node, round(elapsed, 5)


def _reachables(G, root, max_dist, max_size):
    # TODO: is the BFS tree deterministic ?
    # TODO: complete exploration of the started level even if max_size exceeded
    visited = {root: 0}
    nbrs = G[root]
    for d in range(max_dist):
        if not nbrs:
            # all nodes are reachable
            return visited
        new_nbrs = {}
        for n in nbrs:
            if n not in visited:
                visited[n] = d + 1
                new_nbrs.update(G[n])
            if len(visited) == max_size:
                # exceed size limit
                return visited
        nbrs = new_nbrs
    # exceed dist limit
    return visited


def comparison_graph_py(arr1, arr2):
    """ Generate comparison graph
    Comparison graph is a modular product of molecule edges
    """
    # timeout is not implemented
    u1, v1, c1 = zip(*arr1)
    u2, v2, c2 = zip(*arr2)
    c1 = np.array(c1, dtype=int)
    c2 = np.array(c2, dtype=int)
    product = nx.Graph()
    c1 = c1[:, np.newaxis]  # transpose
    if NUMEXPR_AVAILABLE:
        m = ne.evaluate("c2 == c1")
    else:
        m = c2 == c1
    edges = []
    for x, y in zip(*np.nonzero(m)):
        edges.append({"u1": u1[x], "v1": v1[x], "u2": u2[y], "v2": v2[y]})
    # Graph.add_edges is expensive. Add adjacency dict manually.
    node = {}
    for e in edges:
        node[(e["u1"], e["u2"])] = {}
        node[(e["v1"], e["v2"])] = {}
    adj = node.copy()
    for e in edges:
        adj[(e["u1"], e["u2"])][(e["v1"], e["v2"])] = {}
        adj[(e["v1"], e["v2"])][(e["u1"], e["u2"])] = {}
    product = nx.Graph()
    product.node = node
    product.adj = adj
    return product


if not CYTHON_AVAILABLE:
    comparison_graph = comparison_graph_py


class McsdrGls(object):
    def __init__(self, arr1, arr2, timeout):
        self.max1 = arr1[1]
        self.max2 = arr2[1]
        self.map1 = arr1[2]
        self.map2 = arr2[2]
        self.max_clique = []
        self.cg_elapsed = 0
        self.cl_elapsed = 0
        if (arr1[0] and arr2[0]):
            cg, elapsed = comparison_graph(arr1[0], arr2[0], timeout=timeout)
            rest = timeout - elapsed
            cliques, elapsed2 = find_cliques(cg, timeout=rest)
            self.max_clique = max(cliques, key=len, default=[])
            self.cg_elapsed = elapsed
            self.cl_elapsed = elapsed2

    def edge_count(self):
        return len(self.max_clique)

    def local_sim(self, digit=3):
        ecnt = self.edge_count()
        try:
            sim = ecnt / (self.max1 + self.max2 - ecnt)
        except ZeroDivisionError:
            # if both arr1 and arr2 have no edges, no common structures.
            sim = 0
        return round(sim, digit)

    def common_struct(self, mol1):
        new_mol = Compound()
        edges = [self.map1[n1] for n1, _ in self.max_clique]
        atoms = set()
        for u, v in edges:
            atoms |= {u, v}
            bond = mol1.bond(u, v)
            new_mol.add_bond(u, v, bond)
        for a in atoms:
            new_mol.add_atom(a, mol1.atom(a))
        return new_mol

    def exec_time(self):
        return {
            "comparison_graph": round(self.cg_elapsed, 5),
            "find_cliques": round(self.cl_elapsed, 5),
            "total": round(self.cg_elapsed + self.cl_elapsed, 5)
        }


def from_array(arr1, arr2, timeout=10):
    return McsdrGls(arr1, arr2, timeout)


def from_mol(mol1, mol2, timeout=10):
    arr1 = comparison_array(mol1)
    arr2 = comparison_array(mol2)
    return from_array(arr1, arr2, timeout)



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

from collections import Counter
import itertools
import pickle


from chorus.model.graphmol import Compound
from chorus.model.atom import atom_number
from chorus import descriptor, remover, topology, wclogp


def clone(mol):
    """ Returns deepcopy of the object.
    Pickle and unpickle is faster than deepcopy.
    TODO: this query function should be renamed to "cloned"
    """
    return pickle.loads(pickle.dumps(mol, protocol=4))


def assign_descriptors(mol):
    topology.recognize(mol)
    topology.minify_ring(mol)
    descriptor.assign_valence(mol)
    descriptor.assign_rotatable(mol)
    descriptor.assign_aromatic(mol)


def make_Hs_implicit(original_mol, keep_stereo=True):
    """Return molecule that explicit hydrogens removed
    TODO: this query function should be renamed to "explicitHs_removed"
    """
    mol = clone(original_mol)
    mol.descriptors.clear()  # Reset descriptor
    to_remove = set()
    for i, nbrs in mol.neighbors_iter():
        if mol.atom(i).symbol == "H":  # do not remove H2
            continue
        for nbr, bond in nbrs.items():
            if mol.atom(nbr).symbol == "H" and \
                    not (bond.order == 1 and bond.type and keep_stereo):
                to_remove.add(nbr)
    for r in to_remove:
        mol.remove_atom(r)
    assign_descriptors(mol)
    return mol


def null_molecule(assign_descs=True):
    mol = Compound()
    if assign_descs:
        assign_descriptors(mol)
    return mol


def largest_graph(mol):
    """Return a molecule which has largest graph in the compound
    Passing single molecule object will results as same as molutil.clone
    """
    mol.require("Valence")
    mol.require("Topology")
    m = clone(mol)  # Avoid modification of original object
    if m.isolated:
        for k in itertools.chain.from_iterable(m.isolated):
            m.remove_atom(k)
    return m


def mw(mol, ndigits=2):
    """Return standard molecular weight
    :param ndigits: number of digits
    """
    mol.require("Valence")
    return round(sum(a.mw() for _, a in mol.atoms_iter()), ndigits)


def mw_wo_sw(mol, ndigits=2):
    """Molecular weight without salt and water
    :param ndigits: number of digits
    """
    cp = clone(mol)  # Avoid modification of original object
    remover.remove_water(cp)
    remover.remove_salt(cp)
    return round(sum(a.mw() for _, a in cp.atoms_iter()), ndigits)


def charge(mol):
    """Total charge"""
    return sum(a.charge for _, a in mol.atoms_iter())


def non_hydrogen_count(mol):
    """Non-hydrogen atom count """
    return sum(1 for _, a in mol.atoms_iter() if a.symbol != "H")


def H_donor_count(mol):
    """Hydrogen bond donor count """
    mol.require("Valence")
    return sum(1 for _, a in mol.atoms_iter() if a.H_donor)


def H_acceptor_count(mol):
    """Hydrogen bond acceptor count """
    mol.require("Valence")
    return sum(1 for _, a in mol.atoms_iter() if a.H_acceptor)


def rotatable_count(mol):
    """Rotatable bond count """
    mol.require("Rotatable")
    return sum(1 for _, _, b in mol.bonds_iter() if b.rotatable)


def rule_of_five_violation(mol):
    """Lipinski's rule of five violation count """
    v = 0
    if mw(mol) > 500:
        v += 1
    if H_donor_count(mol) > 5:
        v += 1
    if H_acceptor_count(mol) > 10:
        v += 1
    try:
        if wclogp.wclogp(mol) > 5:
            v += 1
    except TypeError:  # N/A
        v += 1
    return v


def composition(mol):
    """Molecular composition in dict format
    (ex. Glucose {'C': 6, 'H': 12, 'O': 6}).
    """
    mol.require("Valence")
    c = Counter()
    for _, a in mol.atoms_iter():
        c += a.composition()
    return c


def available_key(mol):
    """Eldest key +1 """
    return max(mol.key_set() | {1})


def mols_iter(mol):
    keyset = mol.key_set()
    for iso in mol.isolated:
        keyset -= set(iso)
        yield iso
    yield list(keyset)


def formula(mol):
    """Chemical formula.
    Atoms should be arranged in order of C, H and other atoms.
    Molecules should be arranged in order of length of formula text.
    """
    mol.require("Valence")
    mol.require("Topology")
    total_cntr = Counter()
    for m in sorted(mols_iter(mol), key=len, reverse=True):
        cntr = Counter()
        for i in m:
            cntr += mol.atom(i).composition()
        text = []
        Cs = cntr.pop("C", 0)
        if Cs:
            text.append("C")
            if Cs > 1:
                text.append(str(Cs))
        Hs = cntr.pop("H", 0)
        if Hs:
            text.append("H")
            if Hs > 1:
                text.append(str(Hs))
        heteros = sorted(cntr.items(), key=lambda x: atom_number(x[0]))
        for k, v in heteros:
            text.append(k)
            if v > 1:
                text.append(str(v))
        total_cntr["".join(text)] += 1
    total = sorted(total_cntr.items(), key=lambda x: len(x[0]), reverse=True)
    total_text = []
    for k, v in total:
        if v > 1:
            total_text.append(str(v) + k)
        else:
            total_text.append(k)
    return ".".join(total_text)



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

""" SMILES parser module (smilessupplier.py)"""

# TODO: make it simple and eager function

import copy
import re
import logging

from chorus.model.atom import Atom
from chorus.model.bond import Bond
from chorus.model.graphmol import Compound
from chorus import molutil


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


def element():
    atom = []
    bond = Bond()

    def flush():
        """ Flash a connection if exist
        Returns:
          (Atom(), Bond())
          if no atom, returns (None, Bond())
            (ring flag with multiple bond like =1)
          if neither atom and bond, returns None
        Raises:
          ValueError: if the atom symbol is wrong or not supported
        """
        nonlocal bond
        a = None
        logtext = []
        if atom:
            symbol = ''.join(atom)
            a = Atom(symbol.capitalize())
            if symbol in ('c', 'n', 'o', 'p', 's'):
                a.pi = 1
                a.aromatic = True
            logtext.append('Symbol:{}'.format(a.symbol))
        else:
            logtext.append('No element')
        if bond:
            logtext.append('Bond:{}'.format(bond.order))
        else:
            logtext.append('No bond')
        logger.debug(', '.join(logtext))
        result = (a, bond)  # copy
        atom.clear()  # init
        bond = Bond()  # init
        return result

    def exec_(token):
        """ Process a token and return an atom and a bond if available.
        Args:
          token: token
        Returns:
          flush()
        """
        nonlocal bond
        result = (None, None)
        if not token:
            return flush()
        if atom and token.upper() in (
                'B', 'C', 'N', 'O', 'P', 'I', 'S', 'F',
                '=', '#', '/', '\\', '.'):
            result = flush()
        if token == '.':
            bond = None
        elif token == '=':
            bond.order = 2
        elif token == '#':
            bond.order = 3
        elif token == '/':
            bond.smiles_cis_trans = 1
        elif token == '\\':
            bond.smiles_cis_trans = -1
        else:
            atom.append(token)
        logger.debug('{} is queued'.format(token))
        return result

    return exec_


def group():
    container = []
    exp = re.compile("([0-9]*)([a-zA-Z][a-z]?)H?[0-9]*([\+\-]*[0-9]*)")

    def charge_sign(text):
        """ Process charge signs.
        + -> 1, - -> -1, +3 -> 3, +++ -> 3
        Args:
          text: charge signs (+, - or number)
        Returns:
          charge count (neg -2, -1, 0, 1, 2 pos)
        """
        result = 0
        for c in text:
            cases = {'-': -1, '+': 1}
            if c in cases:
                result += cases[c]
            elif c.isdigit():
                result *= int(c)
        return result

    def flush():
        """ Flash an atom group
        explicit hydrogen on the chiral center is added if exist.
        Returns:
          (Atom, hydrogen(as Compound branch))
        Raises:
          ValueError: if the atom symbol is wrong or not supported
        """
        contents = ''.join(container)
        sym = ''
        stereo = 0
        atom = None
        hydro = None
        if '@@' in contents:
            sym = '@@'
            stereo = -1
        elif '@' in contents:
            sym = '@'
            stereo = 1
        if sym:
            # chiral
            c, h = contents.split(sym)
            atom = Atom(c.capitalize())
            if c in ('c', 'n', 'o', 'p', 's'):
                atom.pi = 1
            atom.stereo = stereo
            if h:
                mol = molecule()
                mol('H')
                hydro = mol(None)
        else:
            # charged, isotope or inorganic atoms
            m = exp.match(contents)
            if m:
                logger.debug('Group parsed: {}'.format(m.groups()))
                isotope = m.group(1)
                atom = Atom(m.group(2).capitalize())
                if m.group(2) in ('c', 'n', 'o', 'p', 's'):
                    atom.pi = 1
                atom.charge = charge_sign(m.group(3))
        container.clear()  # init
        return atom, hydro

    def exec_(token):
        """ Process a token and return a group if available.
        Args:
          token: token
        Returns:
          flush()
        """
        if not token:
            return flush()
        container.append(token)

    return exec_


def molecule():
    idx = 0
    depth = 0
    is_group = False
    cp = Compound()
    cp.add_atom(idx, Atom('Dummy'))
    elem = element()
    grp = group()
    branches = []
    ring_labels = {}

    def connect(e):
        """ connect a bond and branches to the molecule
        Args:
          e: (Atom(), Bond())
        """
        nonlocal idx
        for b, r in branches:
            ai = cp.available_idx()
            bond = b.bond(0, 1)
            b.graph.remove_node(0)
            cp.add_molecule(b, bond, idx, 1)
            for k, v in r.items():
                i, bond = v
                ring(k, bond, i + ai - 1)
        logger.debug('{} branches are merged'.format(len(branches)))
        branches.clear()
        if e[0]:
            ai = cp.available_idx()
            cp.add_atom(ai, e[0])
            if e[1]:
                cp.add_bond(idx, ai, e[1])
                logger.debug('Connected:{} {}'.format(e[0].symbol, e[1].order))
            else:
                logger.debug('Added:{}'.format(e[0].symbol))
            idx = ai
        logger.debug(cp)

    def ring(label, bond=None, i=None):
        """ add ring label to the ring_labels dict.
        if the label matched to existing label, flush and close ring
        Args:
          label: number (string type)
          bond: Bond() to close the ring
          i: index (for branch merging)
        """
        if not i:
            i = idx
        if label in ring_labels:
            u = ring_labels[label][0]
            if not bond:
                bond = ring_labels[label][1]
                if not bond:
                    bond = Bond()
            cp.add_bond(u, i, bond)
            del ring_labels[label]
            logger.debug('ring closed')
            logger.debug(ring_labels)
        else:
            ring_labels[label] = (i, bond)
            logger.debug('ring opened')
            logger.debug(ring_labels)

    def flush():
        """ Flash a molecule
        Returns:
          (Compound, ring flag dict)
        """
        nonlocal idx
        logger.debug('Flushed ({} atoms)'.format(cp.atom_count() - 1))
        logger.debug('Ring labels: {}'.format(ring_labels))
        logger.debug(cp)
        result = (copy.deepcopy(cp), copy.deepcopy(ring_labels))  # copy
        idx = 0  # init
        cp.clear()  # init
        cp.add_atom(idx, Atom('Dummy'))  # init
        ring_labels.clear()  # init
        return result

    def exec_(token):
        """ Process a token and return a compound if available.
        Args:
          token: token
        Returns:
          flush()
        """
        nonlocal depth
        nonlocal is_group
        if not token:
            connect(elem(None))
            return flush()
        logger.debug('{} is supplied'.format(token))
        if token == ':':
            return
        elif token == ')':
            depth -= 1
            if depth > 0:
                branches[-1](token)
                logger.debug('Depth {}: {} is sent to branch'.format(depth, token))
            elif not depth:
                branches[-1] = branches[-1](None)
                logger.debug('branch closed'.format(depth))
            else:
                raise ValueError('Syntax Error: unexpected symbol \")\"')
        elif token == '(':
            if depth:
                branches[-1](token)
                logger.debug('Depth {}: {} is sent to branch'.format(depth, token))
            else:
                connect(elem(None))
                branches.append(molecule())
                logger.debug('Depth {}: branch started'.format(depth))
            depth += 1
        else:
            if depth:
                branches[-1](token)
                logger.debug('Depth {}: {} is sent to branch'.format(depth, token))
            elif token == ']':
                g, h = grp(None)
                e = elem(None)
                if e[0]:
                    connect(e)
                    e = (None, Bond())
                connect((g, e[1]))
                if h:
                    branches.append(h)
                is_group = False
            elif is_group:
                    grp(token)
            elif token == '[':
                    is_group = True
            elif token.isdigit():
                e = elem(None)
                if e[0] is None:
                    ring(token, e[1])
                else:
                    connect(e)
                    ring(token)
            else:
                connect(elem(token))

    return exec_


def smiles_to_compound(smiles, assign_descriptors=True):
    """Convert SMILES text to compound object

    Raises:
        ValueError: SMILES with unsupported format
    """
    it = iter(smiles)
    mol = molecule()
    try:
        for token in it:
                mol(token)
        result, _ = mol(None)
    except KeyError as err:
        raise ValueError("Unsupported Symbol: {}".format(err))
    result.graph.remove_node(0)
    logger.debug(result)
    if assign_descriptors:
        molutil.assign_descriptors(result)
    return result



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

class Drawable(object):
    def draw_line(self, p1, p2, c1=(0, 0, 0), c2=None):
        """Draw line segment for bond.

        Args:
            p1 (x, y): the start point
            p2 (x, y): the end point
            c1 (R, G, B): RGB color(0-255) of p1 side half
            c2 (R, G, B): RGB color(0-255) of p2 side half.
                if c2 is undefined, c1 color will be used instead.
        """
        raise NotImplementedError()

    def draw_dashed_line(self, p1, p2, c1=(0, 0, 0), c2=None):
        """Draw dashed line segment for stereobond.

        Args:
            p1 (x, y): the start point
            p2 (x, y): the end point
            c1 (R, G, B): RGB color(0-255) of p1 side half
            c2 (R, G, B): RGB color(0-255) of p2 side half.
                if c2 is undefined, c1 color will be used instead.
        """
        raise NotImplementedError()

    def draw_wave_line(self, p1, p2, color=(0, 0, 0)):
        """Draw wave line segment for stereobond.

        Args:
            p1 (x, y): the start point
            p2 (x, y): the end point
            color (R, G, B): RGB color(0-255)
        """
        raise NotImplementedError()

    def draw_wedge(self, head, tail, color=(0, 0, 0)):
        """Draw wedge (isoscales triangle) for stereobond.

        Args:
            head (x, y): apex of the triangle
            tail (x, y): midpoint of the triangle base
            width (float): triangle base width
            color (R, G, B): RGB color(0-255)
        """
        raise NotImplementedError()

    def draw_dashed_wedge(self, head, tail, color=(0, 0, 0)):
        """Draw dashed wedge (isoscales triangle) for stereobond.

        Args:
            head (x, y): apex of the triangle
            tail (x, y): midpoint of the triangle base
            width (float): triangle base width
            color (R, G, B): RGB color(0-255)
        """
        raise NotImplementedError()

    def draw_text(self, pos, text, color=(0, 0, 0), align="center"):
        """Draw text for atom symbol.

        Args:
            pos (x, y): position of the text center
            text (str): contents
            color (R, G, B): RGB color(0-255)
            align ("right", "center" or "left"): text anchor position
        """
        raise NotImplementedError()



# cython: boundscheck=False, wraparound=False, langage_level=3, profile=True

#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

import cython
import networkx as nx

from libcpp.vector cimport vector
from libc.time cimport clock_t, clock, CLOCKS_PER_SEC


cdef struct Edge:
    int u1
    int v1
    int u2
    int v2


@cython.cdivision(True)
def comparison_graph(arr1, arr2, double timeout):
    cdef clock_t t0 = clock()
    cdef clock_t expire = t0 + <clock_t>(timeout * CLOCKS_PER_SEC)
    u1a, v1a, c1a = zip(*arr1)
    u2a, v2a, c2a = zip(*arr2)
    cdef vector[int] u1 = u1a
    cdef vector[int] v1 = v1a
    cdef vector[long long] c1 = c1a
    cdef vector[int] u2 = u2a
    cdef vector[int] v2 = v2a
    cdef vector[long long] c2 = c2a
    cdef int i, j
    cdef int i_max = c1.size()
    cdef int j_max = c2.size()
    cdef Edge e
    cdef vector[Edge] edges
    product = nx.Graph()
    cdef int interrupted = 0
    for i in range(i_max):
        for j in range(j_max):
            if c1[i] == c2[j]:
                e.u1 = u1[i]
                e.v1 = v1[i]
                e.u2 = u2[j]
                e.v2 = v2[j]
                edges.push_back(e)
                if clock() >= expire:
                    interrupted = 1
                    break
    if not interrupted:
        # Graph.add_edges is expensive. Add adjacency dict manually.
        node = {}
        for edge in edges:
            node[(edge.u1, edge.u2)] = {}
            node[(edge.v1, edge.v2)] = {}
        adj = node.copy()
        for edge in edges:
            adj[(edge.u1, edge.u2)][(edge.v1, edge.v2)] = {}
            adj[(edge.v1, edge.v2)][(edge.u1, edge.u2)] = {}
        product._node = node
        product._adj = adj
    cdef double elapsed = <double>(clock() - t0) / CLOCKS_PER_SEC
    return product, elapsed

"""
def find_cliques_old(graph, timeout=None):
    cdef float t0
    if timeout is not None:
        t0 = time()
    # tuple index may expensive. convert to integer
    cdef int i
    decode = {i: node for i, node in enumerate(graph)}
    encode = {node: i for i, node in decode.items()}
    adj = {}
    for node, adjs in graph.adjacency():
        eadjs = set(encode[a] for a in adjs)
        adj[encode[node]] = eadjs
    # initialize
    result = []
    R = []
    P = set(adj.keys())
    X = set()
    stack = []
    Pv = P - pivot(P, len(P) - 1, adj)
    # DFS loop
    cdef int n, Pcnt, pvdeg
    cdef float now
    while True:
        if timeout is not None:
            now = time.time()
            if now - t0 > timeout:
                return result, False
        if Pv:
            n = Pv.pop()
        else:
            if stack:
                P, X, Pv = stack.pop()
            else:
                break
            R.pop()
            continue
        P.remove(n)
        X.add(n)
        nbrs = adj[n]
        new_P = P & nbrs
        new_X = X & nbrs
        Pcnt = len(new_P)
        if Pcnt in (0, 1):
            if not new_X:
                if Pcnt:
                    result.append([decode[r] for r in R + [n] + list(new_P)])
                else:
                    result.append([decode[r] for r in R + [n]])
            continue
        pvnbrs = pivot(new_X, Pcnt, adj, new_P)
        pvdeg = len(pvnbrs)
        if pvdeg == Pcnt:
            continue
        pvPnbrs = pivot(new_P, Pcnt - 1, adj)
        if len(pvPnbrs) > pvdeg:
            pvnbrs = pvPnbrs
        stack.append((P, X, Pv))
        R.append(n)
        P = new_P
        X = new_X
        Pv = P - pvnbrs
    return result, True


cdef pivot(vs, int goal, adj, filter_=None):
    cdef int max_num = -1
    cdef int d
    cdef int v
    for v in vs:
        if filter is not None:
            a = adj[v]
        else:
            a = adj[v] & filter_
        d = len(a)
        if d > max_num:
            res = a
            max_num = d
            if d == goal:  # short cut
                break
    if max_num == -1:
        return set()
    return res
"""

@cython.cdivision(True)
def find_cliques(G, double timeout):
    # Based on networkX v2.1
    # TODO: slower???
    result = []
    if len(G) == 0:
        return result, 0
    cdef clock_t t0 = clock()
    cdef clock_t expire = t0 + <clock_t>(timeout * CLOCKS_PER_SEC)
    cdef int i, u, q, l
    # tuple index may expensive. convert to integer
    decode = {i: node for i, node in enumerate(G)}
    encode = {node: i for i, node in decode.items()}
    adj = {}
    for node, adjs in G.adjacency():
        eadjs = set()
        for a in adjs:
            eadjs.add(encode[a])
        adj[encode[node]] = eadjs
    Q = [None]
    subg = set(decode)
    cand = set(decode)
    u = max_adj(subg, cand, adj)
    ext_u = cand - adj[u]
    stack = []
    cdef double elapsed
    while True:
        if clock() >= expire:
            break
        if not ext_u:
            Q.pop()
            if not stack:
                break
            subg, cand, ext_u = stack.pop()
            continue
        q = ext_u.pop()
        cand.remove(q)
        Q[len(Q) - 1] = q
        adj_q = adj[q]
        subg_q = subg & adj_q
        if not subg_q:
            result.append([decode[r] for r in Q])
            continue
        cand_q = cand & adj_q
        if cand_q:
            stack.append((subg, cand, ext_u))
            Q.append(None)
            subg = subg_q
            cand = cand_q
            u = max_adj(subg, cand, adj)
            ext_u = cand - adj[u]
    elapsed = <double>(clock() - t0) / CLOCKS_PER_SEC
    return result, elapsed


cdef max_adj(subg, cand, adj):
    cdef int maxadj = -1
    cdef int s, numadj, res
    for s in subg:
        numadj = len(cand & adj[s])
        if numadj > maxadj:
            maxadj = numadj
            res = s
    return res



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

import os
import yaml
from collections import Counter

""" Import periodic table """
with open(os.path.join(
        os.path.dirname(__file__), "periodictable.yaml")) as file:
    P_TAB = yaml.load(file.read())


def atom_number(symbol):
    return P_TAB[symbol]['number']


ALIASES = {
    "name": "name",
    "n": "number",
    "sym": "symbol",
    "chg": "charge",
    "hs": "H_count",
    "vis": "visible",
    "crds": "coords",
    "c": "color",
    "pi": "pi",
    "arom": "aromatic",
    "hdo": "H_donor",
    "hac": "H_acceptor",
    "cbnl": "carbonyl_C",
    "lp": "lone_pair",
    "wc": "wctype",
    "patty": "patty",
    "stereo": "stereo"
}


class Atom(object):
    """Atom object

    Parameters:
        symbol (str): atom symbol
        attr_dict (dict): load ``Atom`` object from dict notation.

    Attributes:
        name: general name of the atom
        number: atomic number
        symbol: atom symbol
        charge: number of electric charges
        H_count: number of hydrogens on the atom
        multi: multiplicity (1: non-radical, 2: radical, 3: biradical)
        mass: molar mass
        visible: whether it should be drawn or not
        coords (tuple): (x, y, z)
        color (tuple): (R, G, B) in the range of 0-255
        pi: number of pi electron
        aromatic: aromatic
        H_donor: Hydrogen bond donor
        H_acceptor: Hydrogen bond acceptor
        wctype: ``wclogp``` descriptor
        patty: ``patty``` descriptor

    """
    def __init__(self, symbol, attr_dict=None):
        self.name = P_TAB[symbol]['name']
        self.number = P_TAB[symbol]['number']
        self.symbol = symbol
        self.charge = 0
        self.H_count = 0
        self.multi = 1
        self.mass = None
        self.visible = 1
        if symbol == 'C':
            self.visible = 0
        self.coords = None
        self.color = tuple(P_TAB[symbol].get('color', [0, 192, 192]))

        # chem.descriptor
        self.pi = 0
        self.aromatic = 0
        self.H_donor = 0
        self.H_acceptor = 0
        self.carbonyl_C = 0
        self.lone_pair = 0  # not used?
        if self.symbol in ('N', 'O'):
            self.lone_pair = 1
        if self.symbol in ('N', 'O', 'F'):
            self.H_acceptor = 1
        self.wctype = None
        self.patty = 7

        # for SMILES
        self.stereo = 0
        # self.excited = False
        # self.stereo_flag = None

        # loads attribute dict
        if attr_dict is not None:
            for k, v in attr_dict.items():
                setattr(self, ALIASES[k], v)

    def __getattr__(self, name):
        if name not in ALIASES:
            raise AttributeError()
        return object.__getattribute__(self, ALIASES[name])

    def add_hydrogen(self, num):
        """Adds hydrogens

        Args:
            num (int): number of hydrogens
        """
        self.H_count = num
        if num > 0 and self.symbol in ("N", "O"):
            self.H_donor = 1
        else:
            self.H_donor = 0

    def formula_html(self, reversed_=False):
        """Chemical formula HTML

        Args:
            reversed (bool): reversed text for leftmost atom groups
        """
        if self.H_count == 1:
            text = "H"
        elif self.H_count > 1:
            text = "H<sub>{}</sub>".format(self.H_count)
        else:
            text = ""
        seq = [self.symbol, text, self.charge_sign_html()]
        if reversed_:
            seq = reversed(seq)
        return "".join(seq)

    def composition(self):
        """Atom group composition

        Returns:
            collections.Counter: atom count by atom symbol
        """
        return Counter({self.symbol: 1, 'H': self.H_count})

    def mw(self):
        """Molecular weight"""
        m = P_TAB[self.symbol].get('std_weight')
        mh = P_TAB['H'].get('std_weight')
        return m + mh * self.H_count

    def charge_sign(self):
        """Charge sign text"""
        if self.charge > 0:
            sign = "+"
        elif self.charge < 0:
            sign = "–"  # en dash, not hyphen-minus
        else:
            return ""
        ab = abs(self.charge)
        if ab > 1:
            return str(ab) + sign
        return sign

    def charge_sign_html(self):
        """Charge sign HTML"""
        if self.charge:
            return "<sup>{}</sup>".format(self.charge_sign())
        return ""

    def dumps(self):
        """Dumps Atom object to dict notation"""
        res = {}
        for k, v in ALIASES.items():
            res[k] = getattr(self, v)
        return res



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

import chorus


def atom_block(mol):
    idx_table = {}
    lines = []
    for i, (k, atom) in enumerate(mol.atoms_iter()):
        atom_line = []
        for ax in atom.coords:
            int_, frac = str(ax).split(".")
            atom_line.append("{:>5}".format(int_))
            atom_line.append("{:0<4}".format(frac))
        atom_line.append("{:<3}".format(atom.symbol))
        lines.append(
            "{}.{}{}.{}{}.{} {} 0  0  0  0  0  0  0  0  0  0  0  0".format(
                *atom_line)
        )
        idx_table[k] = i + 1
    return lines, idx_table


def bond_block(mol, idx_table):
    stereo_conv = {
        1: {0: 0, 1: 1, 3: 4, 2: 6},
        2: {0: 0, 1: 1, 3: 3, 2: 6}
    }
    lines = []
    for u, v, b in mol.bonds_iter():
        bond_line = []
        bond_line.append("{:>3}".format(idx_table[u]))
        bond_line.append("{:>3}".format(idx_table[v]))
        bond_line.append("{:>3}".format(b.order))
        try:
            bond_line.append("{:>3}".format(stereo_conv[b.order][b.type]))
        except KeyError:
            bond_line.append("{:>3}".format(0))
        lines.append("{}{}{}{}  0  0  0".format(*bond_line))
    return lines


def prop_block(mol, idx_table):
    lines = []
    chg_line = []
    rad_line = []
    iso_line = []
    for k, atom in mol.atoms_iter():
        idx = "{:>3}".format(idx_table[k])
        if atom.charge:
            chg = "{:>3}".format(atom.charge)
            chg_line.append(" {} {}".format(idx, chg))
        if atom.multi != 1:
            rad = "{:>3}".format(atom.multi)
            rad_line.append(" {} {}".format(idx, rad))
        if atom.mass is not None:
            iso = "{:>3}".format(atom.mass)
            iso_line.append(" {} {}".format(idx, iso))
    if len(chg_line):
        chg_count = "{:>3}".format(len(chg_line))
        lines.append("M  CHG{}{}".format(chg_count, "".join(chg_line)))
    if len(rad_line):
        rad_count = "{:>3}".format(len(rad_line))
        lines.append("M  RAD{}{}".format(rad_count, "".join(rad_line)))
    if len(iso_line):
        iso_count = "{:>3}".format(len(iso_line))
        lines.append("M  ISO{}{}".format(iso_count, "".join(iso_line)))
    return lines


def data_block(mol):
    lines = []
    for k, v in mol.data.items():
        lines.append("> <{}>".format(k))
        lines.append(str(v))
        lines.append("")
    lines.append("$$$$")
    return lines


def mol_block(mol, sdfile=True):
    lines = [
        "",
        "Chorus version {}".format(chorus.VERSION),
        ""
    ]
    chiral_flag = 0
    count_line = [
        "{:>3}".format(mol.atom_count()),
        "{:>3}".format(mol.bond_count()),
        "{:>3}".format(chiral_flag)
    ]
    lines.append("{}{}  0  0{}  0  0  0  0  0999 V2000".format(*count_line))
    atoms, idx_table = atom_block(mol)
    lines.extend(atoms)
    bonds = bond_block(mol, idx_table)
    if bonds:
        lines.extend(bonds)
    props = prop_block(mol, idx_table)
    if props:
        lines.extend(props)
    lines.append("M  END")
    if sdfile:
        lines.extend(data_block(mol))
    lines.append("")
    return "\n".join(lines)


def mols_to_text(mols):
    """Converts molecules to the SDFile format text

    Args:
        mols: list of molecule objects

    Returns:
        SDFile text
    """
    return "".join(mol_block(mol) for mol in mols)


def mols_to_file(mols, path):
    """Save molecules to the SDFile format file

    Args:
        mols: list of molecule objects
        path: file path to save
    """
    with open(path, 'w') as f:
        f.write(mols_to_text(mols))



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

import base64
import functools
import math

from chorus.draw import drawable, drawer2d, helper
from chorus.util.geometry import scale, distance, m_seg, p_seg
from chorus.util.text import substitute

FONT_SIZE = 14   # Font size(pixel)


class SVG(drawable.Drawable):
    """Draw chemical structure in SVG format.

    Args:
        molecule: compound
        screen_size (width, height): pixel size to display

    Attributes:
        font_weight (str): font weight
        font_family (str): font family
        screen_size (width, height): pixel size to display
        logical_size (width, height): pixel size of actual viewport
    """
    def __init__(self, molecule, screen_size=(180, 180)):
        mol = helper.ready_to_draw(molecule)
        self.font_weight = "normal"
        self.font_family = "Helvetica"
        self.bgcolor = None  # background fill (if None, transparent)
        self.displaymlb = 30  # suitable for FONT_SIZE=14 and default linewidth
        self.margin = self.displaymlb
        w, h, mlb = mol.size2d
        self.scale_factor = self.displaymlb / mlb
        self.original_size = (w, h)  # for coords conversion
        self.screen_size = screen_size  # svg object display size
        self._header = ['<svg xmlns="http://www.w3.org/2000/svg"']
        self._header.append(' xmlns:xlink="http://www.w3.org/1999/xlink"')
        self._header.append(' version="1.2" baseProfile="tiny"')
        self._header.append(' text-rendering="geometricPrecision"')
        self._elems = []
        drawer2d.draw(self, mol)

    def contents(self):
        """Get svg string
        """
        c = self._header[:]
        c.append(' font-weight="{}"'.format(self.font_weight))
        c.append(' font-family="{}"'.format(self.font_family))
        c.append(' width="{}" height="{}"'.format(*self.screen_size))
        sclw = self.original_size[0] * self.scale_factor
        sclh = self.original_size[1] * self.scale_factor
        longside = max([sclw, sclh])
        width = round(longside + self.margin * 2, 2)
        height = round(longside + self.margin * 2, 2)
        xleft = round(-self.margin - (longside - sclw) / 2, 2)
        ytop = round(-self.margin - (longside - sclh) / 2, 2)
        c.append(' viewBox="{} {} {} {}">\n'.format(
            xleft, ytop, width, height))
        if self.bgcolor is not None:
            c.append('<rect x="{}", y="{}" width="{}" height="{}" fill="{}" \
                />\n'.format(xleft, ytop, width, height, self.bgcolor))
        c.extend(self._elems)
        c.append("</svg>")
        return "".join(c)

    def data_url_scheme(self):
        """Get svg in Data URL Scheme format.
        """
        # TODO: move to web.app or make it function
        # remove #svg from dataframe
        encoded = base64.b64encode(self.contents().encode())
        return "data:image/svg+xml;base64," + encoded.decode()

    def save(self, path):
        """Save svg as file(.svg)

        Args:
            path (str): destination to save file
        """
        with open(path, 'w') as f:
            f.write(self.contents())

    def _coords_conv(self, pos):
        """For Svg coordinate system, reflect over X axis and
        translate from center to top-left
        """
        px = (self.original_size[0] / 2 + pos[0]) * self.scale_factor
        py = (self.original_size[1] / 2 - pos[1]) * self.scale_factor
        return round(px, 2), round(py, 2)

    def line_conv(method):
        @functools.wraps(method)
        def wrapper(self, p1, p2, c1, c2=None):
            p1_ = self._coords_conv(p1)
            p2_ = self._coords_conv(p2)
            # rgb(255,255,255)
            c1_ = 'rgb({},{},{})'.format(*c1)
            if c2:
                c2_ = 'rgb({},{},{})'.format(*c2)
                method(self, p1_, p2_, c1_, c2_)
            else:
                method(self, p1_, p2_, c1_)
        return wrapper

    def text_conv(method):
        @functools.wraps(method)
        def wrapper(self, p, text, c, align):
            p_ = self._coords_conv(p)
            c_ = 'rgb({},{},{})'.format(*c)
            method(self, p_, text, c_, align)
        return wrapper

    def _draw_line_base(self, p1, p2, c1, c2=None, op=""):
        tmpl = '<line x1="{}" y1="{}" x2="{}" y2="{}" stroke="{}"{} />\n'
        if c2 and c2 != c1:
            pmx, pmy = (
                round((p1[0] + p2[0]) / 2, 2),
                round((p1[1] + p2[1]) / 2, 2)
            )
            self._elems.append(tmpl.format(p1[0], p1[1], pmx, pmy, c1, op))
            self._elems.append(tmpl.format(pmx, pmy, p2[0], p2[1], c2, op))
        else:
            self._elems.append(tmpl.format(p1[0], p1[1], p2[0], p2[1], c1, op))

    @line_conv
    def draw_line(self, p1, p2, c1, c2=None):
        self._draw_line_base(p1, p2, c1, c2)

    @line_conv
    def draw_dashed_line(self, p1, p2, c1, c2=None):
        op = ' stroke-dasharray="10,10"'
        self._draw_line_base(p1, p2, c1, c2, op)

    @line_conv
    def draw_wedge(self, head, tail, color):
        width = self.displaymlb * 0.2
        tmpl = '<polygon points="{}" fill="{}" />\n'
        t1 = m_seg(tail, head, math.pi / -2, width / 2)[0]
        t2 = m_seg(tail, head, math.pi / 2, width / 2)[0]
        ptext = ' '.join(['{},{}'.format(
            round(x, 2), round(y, 2)) for x, y in (head, t1, t2)])
        self._elems.append(tmpl.format(ptext, color))

    @line_conv
    def draw_dashed_wedge(self, head, tail, color):
        width = self.displaymlb * 0.2
        tmpl = '<line x1="{}" y1="{}" x2="{}" y2="{}" stroke="{}" />\n'
        t1 = m_seg(tail, head, math.pi / -2, width / 2)[0]
        t2 = m_seg(tail, head, math.pi / 2, width / 2)[0]
        step_num = 7
        for i in range(step_num):
            step = distance(head, tail) / step_num * i
            trim = 1 / step_num * i
            tp1, tp2 = p_seg(t1, t2, True, step, trim)
            self._elems.append(
                tmpl.format(round(tp1[0], 2), round(tp1[1], 2),
                            round(tp2[0], 2), round(tp2[1], 2), color))

    @line_conv
    def draw_wave_line(self, p1, p2, color):
        width = self.displaymlb * 0.2
        tmpl = '<polyline points="{}" stroke="{}" fill="none" />\n'
        s1 = m_seg(p1, p2, math.pi / -2, width / 2)
        s2 = m_seg(p1, p2, math.pi / 2, width / 2)
        ps = [p1]
        for i, s in enumerate((s1, s2, s1, s2, s1, s2)):
            p = scale(s[1], (i + 1) / 7, s[0])
            ps.append(p)
        ps.append(p2)
        ptext = ' '.join(['{},{}'.format(
            round(x, 2), round(y, 2)) for x, y in ps])
        self._elems.append(tmpl.format(ptext, color))

    @text_conv
    def draw_text(self, pos, text, color, align="center"):
        tmpl = '<text x="{}" y="{}" font-size="{}" fill="{}"{}>{}</text>\n'
        stmpl = '<tspan baseline-shift="{}" font-size="{}">'
        rt = substitute({
            '<sub>': stmpl.format("-25%", FONT_SIZE * 0.7),
            '<sup>': stmpl.format("50%", FONT_SIZE * 0.7),
            '</sub>': '</tspan>',
            '</sup>': '</tspan>'
        }, text)
        op = {"center": ' text-anchor="middle"',
              "right": ' text-anchor="end"', "left": ''}
        xoffset = {"center": 0, "right": FONT_SIZE / 2,
                   "left": -FONT_SIZE / 2}
        px = pos[0] + xoffset[align]
        py = pos[1] + FONT_SIZE / 2
        self._elems.append(
            tmpl.format(round(px, 2), round(py, 2),
                        FONT_SIZE, color, op[align], rt))


def mol_to_svg(mol, screen_size=(180, 180)):
    """Convenient method to generate SVG image"""
    return SVG(mol, screen_size).contents()



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

import operator
from collections import Counter

from networkx.algorithms.isomorphism.vf2userfunc import GraphMatcher

from chorus import molutil


def atom_match(n1, n2):
    a1 = n1['atom']
    a2 = n2['atom']
    if a1.symbol == a2.symbol and a1.pi == a2.pi:
        return True


def filter_(mol, query, f=operator.ne):
    mol.require("Topology")
    query.require("Topology")
    if f(molutil.composition(mol), molutil.composition(query)):
        return False
    if f(Counter(a.pi for _, a in mol.atoms_iter()),
         Counter(a.pi for _, a in query.atoms_iter())):
        return False
    if f(Counter(len(r) for r in mol.rings),
         Counter(len(r) for r in query.rings)):
        return False
    if f(Counter(len(s) for s in mol.scaffolds),
         Counter(len(s) for s in query.scaffolds)):
        return False
    return True


def equal(mol, query, largest_only=True, ignore_hydrogen=True):
    """ if mol is exactly same structure as the query, return True
    Args:
      mol: Compound
      query: Compound
    """
    m = molutil.clone(mol)
    q = molutil.clone(query)
    if largest_only:
        m = molutil.largest_graph(m)
        q = molutil.largest_graph(q)
    if ignore_hydrogen:
        m = molutil.make_Hs_implicit(m)
        q = molutil.make_Hs_implicit(q)
    if molutil.mw(m) == molutil.mw(q):
        gm = GraphMatcher(q.graph, m.graph, node_match=atom_match)
        return gm.is_isomorphic()
    return False


def substructure(mol, query, largest_only=True, ignore_hydrogen=True):
    """ if mol is a substructure of the query, return True
    Args:
      mol: Compound
      query: Compound
      largest_only: compare only largest graph molecule
    """
    def subset_filter(cnt1, cnt2):
        diff = cnt2
        diff.subtract(cnt1)
        if any(v < 0 for v in diff.values()):
            return True

    if not (len(mol) and len(query)):
        return False  # two blank molecules are not isomorphic
    m = molutil.clone(mol)
    q = molutil.clone(query)
    if largest_only:
        m = molutil.largest_graph(m)
        q = molutil.largest_graph(q)
    if ignore_hydrogen:
        m = molutil.make_Hs_implicit(m)
        q = molutil.make_Hs_implicit(q)
    if filter_(m, q, f=subset_filter):
        gm = GraphMatcher(q.graph, m.graph, node_match=atom_match)
        return gm.subgraph_is_isomorphic()
    return False


def superstructure(mol, query, largest_only=True, ignore_hydrogen=True):
    return substructure(query, mol, largest_only, ignore_hydrogen)



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

from chorus.util import iterator


def assign_valence(mol):
    """Assign pi electron and hydrogens"""
    for u, v, bond in mol.bonds_iter():
        if bond.order == 2:
            mol.atom(u).pi = 1
            mol.atom(v).pi = 1
            if mol.atom(u).symbol == "O" and not mol.atom(u).charge:
                mol.atom(v).carbonyl_C = 1
            if mol.atom(v).symbol == "O" and not mol.atom(v).charge:
                mol.atom(u).carbonyl_C = 1
        elif bond.order == 3:
            mol.atom(u).pi = mol.atom(v).pi = 2
    max_nbr = {"C": 4, "Si": 4, "N": 3, "P": 3, "As": 3,
               "O": 2, "S": 2, "Se": 2, "F": 1, "Cl": 1, "Br": 1, "I": 1}
    for i, nbrs in mol.neighbors_iter():
        atom = mol.atom(i)
        if len(nbrs) == 2 and all(bond.order == 2 for bond in nbrs.values()):
            atom.pi = 2  # sp (allene, ketene)
        if atom.symbol in max_nbr:
            h_cnt = max_nbr[atom.symbol] - len(nbrs) - atom.pi + atom.charge
            if h_cnt > 0:
                mol.atom(i).add_hydrogen(h_cnt)
    mol.descriptors.add("Valence")


def assign_rotatable(mol):
    """Assign rotatable bonds (single and not terminal and not in ring)"""
    mol.require("Valence")
    mol.require("Topology")
    bs = set([(u, v) for u, v, _ in mol.bonds_iter()])
    for r in mol.rings:
        for a, b in iterator.consecutive(r + [r[0]], 2):
            bs -= {(a, b)}
            bs -= {(b, a)}
    for u, v in bs:
        unc = mol.neighbor_count(u)
        vnc = mol.neighbor_count(v)
        bc = mol.bond(u, v).order
        if bc == 1 and (unc > 1 and vnc > 1):
            mol.bond(u, v).rotatable = True
    mol.descriptors.add("Rotatable")


def assign_aromatic(mol):
    """Assign aromatic ring
    sp2 atom:
    pi=1 -> +1
    N, O, S, C- -> +2
    >C=O, B, C+ -> 0
    sp3 atom -> not aromatic
    sum of the score satisfies 4n+2 -> aromatic
    """
    mol.require("Valence")
    mol.require("MinifiedRing")
    for ring in mol.rings:
        pi_cnt = 0
        for r in ring:
            if mol.atom(r).pi == 0:
                if mol.atom(r).symbol == "C":
                    if mol.atom(r).charge == 1:
                        pass
                    elif mol.atom(r).charge == -1:
                        pi_cnt += 2
                    else:
                        break
                elif mol.atom(r).charge == 0:
                    if mol.atom(r).symbol in ("N", "O", "S"):
                        pi_cnt += 2
                    elif mol.atom(r).symbol == "B":
                        pass
                    else:
                        break
                else:
                    break
            elif mol.atom(r).pi == 1:
                if mol.atom(r).carbonyl_C:
                    pass
                else:
                    pi_cnt += 1
            else:
                break
        else:
            if pi_cnt % 4 == 2:
                for u, v in iterator.consecutive(ring + [ring[0]], 2):
                    mol.atom(u).aromatic = True
                    mol.bond(u, v).aromatic = True
    mol.descriptors.add("Aromatic")


def assign_charge(mol, force_recalc=False):
    """Assign charges in physiological condition"""
    # TODO: not implemented yet
    mol.require("Aromatic")
    for i, nbrs in mol.neighbors_iter():
        atom = mol.atom(i)
        nbrcnt = len(nbrs)
        if atom.symbol == "N":
            if not atom.pi:
                # non-conjugated amines are anion
                mol.atom(i).charge_phys = 1
            elif nbrcnt == 1 and atom.pi == 2:
                # amidine, guanidine are conjugated cation
                ni = list(nbrs.keys())[0]
                conj = False
                sp2n = None
                for nni, nnb in mol.neighbors(ni).items():
                    if mol.atom(nni).symbol == "N" and nnb.order == 2 \
                            and not mol.atom(nni).aromatic:
                        mol.atom(nni).charge_conj = 1
                        conj = True
                    elif mol.atom(nni).symbol == "N" and nni != i:
                        sp2n = nni
                if conj:
                    mol.atom(i).charge_phys = 1
                    if sp2n is not None:
                        mol.atom(sp2n).charge_conj = 1
        elif atom.symbol == "O" and nbrcnt == 1 and atom.pi == 2:
            # oxoacid are conjugated anion
            ni = list(nbrs.keys())[0]
            conj = False
            if mol.atom(ni).symbol == "N":
                mol.atom(i).n_oxide = True
                mol.atom(ni).n_oxide = True
            for nni, nnb in mol.neighbors(ni).items():
                if mol.atom(nni).symbol in ("O", "S") \
                        and nnb.order == 2 and not mol.atom(ni).n_oxide:
                    mol.atom(nni).charge_conj = -1
                    conj = True
            if conj:
                mol.atom(i).charge_phys = -1
        elif atom.symbol == "S" and nbrcnt == 1:
            # thiophenols are anion
            ni = list(nbrs.keys())[0]
            if mol.atom(ni).aromatic:
                mol.atom(i).charge_phys = -1
    mol.charge_assigned = True
    mol.descriptors.add("Phys_charge")



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

ALIASES = {
    "o": "order",
    "vis": "visible",
    "arom": "aromatic",
    "rot": "rotatable",
    "ilf": "is_lower_first",
    "type": "type",
    "sgeo": "smiles_cis_trans"
}


class Bond(object):
    """Bond object

    * Bond type
        * Single bond
            * 0: first - second
            * 1: first ◀ second (Up-arrow)
            * 2: first ◁ second (Down-arrow)
            * 3: first ~ second (Chiral)

        *Double bond
            * 0: second ニ first (clockwise, default)
            * 1: first ニ second (counter-clockwise)
            * 2: first ＝ second (equal length, for terminal bond by default)
            * 3: first × second (Cis-Trans Unknown)

    Parameters:
        attr_dict (dict): load ``Bond`` object from dict type notation.

    Attributes:
        order (int): bond order. ``smilesparser`` assigns order=9 to
            aromatic bonds.
        is_lower_first (bool): bond direction. If true, lower index atom is
            the first atom and the other is the second atom. ``Bond.type``
            attribute uses this feature for stereochemistry determination.
        type: bond type (see above)
        rotatable (bool): rotatable or not
        aromatic (bool): aromatic or not
        smiles_cis_trans: SMILES cis/trans flag
        visible (bool): whether it should be drawn or not.


    """
    def __init__(self, attr_dict=None):
        self.order = 1
        self.is_lower_first = 1
        self.type = 0
        self.rotatable = 0
        self.aromatic = 0
        self.smiles_cis_trans = 0
        self.visible = 1

        # loads attribute dict
        if attr_dict is not None:
            for k, v in attr_dict.items():
                setattr(self, ALIASES[k], v)

    def __getattr__(self, name):
        if name not in ALIASES:
            raise AttributeError()
        return object.__getattribute__(self, ALIASES[name])

    def dumps(self):
        res = {}
        for k, v in ALIASES.items():
            res[k] = getattr(self, v)
        return res



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

import functools
import io
import math

import matplotlib as mpl
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.font_manager import FontProperties

from chorus.draw import drawable, drawer2d, helper
from chorus.util.geometry import scale, distance, m_seg, p_seg
from chorus.util.text import substitute

mpl.rcParams["mathtext.default"] = "regular"
mpl.rcParams["font.sans-serif"] = [
    "Arial", "Helvetica", "Liberation Sans", "Bitstream Vera Sans"]


def basefont(size):
    f = FontProperties()
    f.set_size(size)
    return f


class Matplotlib(drawable.Drawable):
    """Draw chemical structure with Matplotlib.

    Args:
        molecule: compound

    Attributes:
        font_weight (str): font weight
        font_family (str): font family
        screen_size (width, height): pixel size to display
        logical_size (width, height): pixel size of actual viewport
    """
    def __init__(self, molecule):
        mol = helper.ready_to_draw(molecule)
        self.dpi = 120
        self.fig = mpl.figure.Figure()
        FigureCanvas(self.fig)
        self.fig.subplots_adjust(left=0, right=1, top=1, bottom=0,
                                 wspace=0, hspace=0)
        self.ax = self.fig.add_subplot(111)
        w, h, mlb = mol.size2d
        self.displaymlb = 60
        self.margin = self.displaymlb
        self.scale_factor = self.displaymlb / mlb
        scaledw = w * self.scale_factor + self.margin * 2
        scaledh = h * self.scale_factor + self.margin * 2
        longside = max([scaledw, scaledh])
        self.font_size = 16
        self.fig.set_size_inches(longside / self.dpi, longside / self.dpi)
        self.ax.set_xlim(-scaledw / 2, scaledw / 2)
        self.ax.set_ylim(-scaledh / 2, scaledh / 2)
        self.ax.set_aspect("equal", "box")
        self.ax.axis('off')
        drawer2d.draw(self, mol)

    def to_bytesio(self):
        buf = io.BytesIO()
        self.fig.savefig(buf, dpi=self.dpi)  # resume scale
        return buf

    def save(self, path):
        self.fig.savefig(path, dpi=self.dpi)  # resume scale

    def get_size(self):
        return tuple(int(s * self.dpi) for s in self.fig.get_size_inches())

    def line_conv(method):
        @functools.wraps(method)
        def wrapper(self, p1, p2, c1, c2=None):
            p1_ = tuple(c * self.scale_factor for c in p1[:2])
            p2_ = tuple(c * self.scale_factor for c in p2[:2])
            # rgb in the range of 0 to 1
            c1_ = tuple(1 - ((255 - c) / 255) for c in c1)
            if c2:
                c2_ = tuple(1 - ((255 - c) / 255) for c in c2)
                method(self, p1_, p2_, c1_, c2_)
            else:
                method(self, p1_, p2_, c1_)
        return wrapper

    def text_conv(method):
        @functools.wraps(method)
        def wrapper(self, p, text, c, align):
            p_ = tuple(c * self.scale_factor for c in p[:2])
            c_ = tuple(1 - ((255 - c) / 255) for c in c)
            method(self, p_, text, c_, align)
        return wrapper

    def _draw_line_base(self, p1, p2, c1, c2=None, ls="solid"):
        if c2 and c2 != c1:
            pm = ((p1[0] + p2[0]) / 2, (p1[1] + p2[1]) / 2)
            px, py = [tuple(p) for p in zip(p1, pm)]
            self.ax.add_line(mpl.lines.Line2D(
                px, py, c=c1, lw=1.5, ls=ls, clip_on=False
            ))
            px, py = [tuple(p) for p in zip(pm, p2)]
            self.ax.add_line(mpl.lines.Line2D(
                px, py, c=c2, lw=1.5, ls=ls, clip_on=False
            ))
        else:
            px, py = [tuple(p) for p in zip(p1, p2)]
            self.ax.add_line(mpl.lines.Line2D(
                px, py, c=c1, lw=1.5, ls=ls, clip_on=False
            ))

    @line_conv
    def draw_line(self, p1, p2, c1, c2=None):
        self._draw_line_base(p1, p2, c1, c2)

    @line_conv
    def draw_dashed_line(self, p1, p2, c1, c2=None):
        self._draw_line_base(p1, p2, c1, c2, ls="dashed")

    @line_conv
    def draw_wedge(self, head, tail, color):
        width = self.displaymlb * 0.2
        self.ax.add_patch(mpl.patches.FancyArrowPatch(
            tail, head, ec=color, fc=color, clip_on=False,
            arrowstyle="wedge,tail_width={},shrink_factor=0.5".format(width)
        ))

    @line_conv
    def draw_dashed_wedge(self, head, tail, color):
        width = self.displaymlb * 0.2
        t1 = m_seg(tail, head, math.pi / -2, width / 2)[0]
        t2 = m_seg(tail, head, math.pi / 2, width / 2)[0]
        step_num = 7
        for i in range(step_num):
            step = distance(head, tail) / step_num * i
            trim = 1 / step_num * i
            tp1, tp2 = p_seg(t1, t2, True, step, trim)
            tpx, tpy = [tuple(p) for p in zip(tp1, tp2)]
            self.ax.add_line(mpl.lines.Line2D(
                tpx, tpy, c=color, lw=1.5, clip_on=False
            ))

    @line_conv
    def draw_wave_line(self, pos1, pos2, color):
        width = self.displaymlb * 0.2
        s1 = m_seg(pos1, pos2, math.pi / -2, width / 2)
        s2 = m_seg(pos1, pos2, math.pi / 2, width / 2)
        codes = [mpl.path.Path.MOVETO]
        vers = [pos1]
        for i, s in enumerate((s1, s2, s1, s2, s1, s2)):
            p = scale(s[1], (i + 1) / 7, s[0])
            codes.append(mpl.path.Path.LINETO)
            vers.append(p)
        codes.append(mpl.path.Path.LINETO)
        vers.append(pos2)
        path = mpl.path.Path(vers, codes)
        self.ax.add_patch(mpl.patches.PathPatch(
            path, fc="None", ec=color, lw=1.5, clip_on=False
        ))

    @text_conv
    def draw_text(self, pos, text, color, align="center"):
        replaced = substitute({
            '<sub>': '$_{',
            '<sup>': '$^{',
            '</sub>': '}$',
            '</sup>': '}$'
        }, text)
        xcorr = pos[0]
        if align == "left":
            xcorr -= self.font_size  # TODO: empirical
        elif align == "right":
            xcorr += self.font_size  # TODO: empirical
        self.ax.text(
            xcorr, pos[1], replaced, color=color, clip_on=False,
            fontproperties=basefont(self.font_size), ha=align, va="center"
        )



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

import io
import math

from PIL import Image
from PySide import QtCore, QtGui, QtSvg

from chorus.draw.drawermixin2d import DrawerMixin2D
from chorus.draw import helper
from chorus.util.geometry import distance, scale, m_seg, p_seg

MIN_LOGICAL_SIZE = 250
MARGIN = 40      # Output image margin
LINE_WEIGHT = 1  # Line weight(pixel)
RGB_TMPL = 'rgb({},{},{})'


class Qt(DrawerMixin2D):
    """Deprecated(no longer updated)"""
    def __init__(self):
        self.screen_size = (250, 250)

        self.contents = None
        self.logical_size = None
        self.painter = None
        # Get GUI instance
        app = QtGui.QApplication.instance()
        if app is None:
            QtGui.QApplication(())

    def draw(self, compound):
        helper.ready_to_draw(compound)
        size = max([compound.size2d[0] + MARGIN,
                    compound.size2d[1] + MARGIN,
                    MIN_LOGICAL_SIZE])
        self.logical_size = (size, size)
        self.contents = QtGui.QPixmap(*self.logical_size)
        self.painter = QtGui.QPainter(self.contents)
        self.painter.setRenderHint(QtGui.QPainter.Antialiasing, True)
        self.painter.setRenderHint(QtGui.QPainter.SmoothPixmapTransform, True)
        self.painter.fillRect(0, 0, size, size, QtCore.Qt.white)
        super().draw(compound)
        self.painter.end()

    def load_svg(self, str_):
        xml = QtCore.QXmlStreamReader(str_)
        svg = QtSvg.QSvgRenderer(xml)
        w, h = svg.viewBoxF().size().toTuple()
        s = max((w, h,  MIN_LOGICAL_SIZE))
        self.contents = QtGui.QPixmap(s, s)
        painter = QtGui.QPainter(self.contents)
        painter.fillRect(0, 0, s, s, QtCore.Qt.white)
        svg.render(painter, QtCore.QRectF((s-w) / 2, (s-h) / 2, w, h))
        painter.end()
        # self.contents = self.contents.scaledToHeight(
        #    self.screen_size[1], QtCore.Qt.SmoothTransformation)

    def save_png_bytes_io(self):
        """ Save contents as PNG format file(BytesIO object)
        Note: DPI is platform dependent due to QPaintDevice DPI handling.
        Mac -> 72dpi, Ubuntu -> 96dpi
        """
        bytearr = QtCore.QByteArray()
        buf = QtCore.QBuffer(bytearr)
        buf.open(QtCore.QIODevice.WriteOnly)
        self.contents.save(buf, "PNG")
        bio = io.BytesIO(bytearr.data())
        # DPI correction
        img = Image.open(bio)
        img = img.resize((self.screen_size[0], self.screen_size[1]),
                         Image.ANTIALIAS)
        img.info["dpi"] = (96, 96)
        converted = io.BytesIO()
        img.save(converted, format='png')
        return converted

    def save_png(self, filename):
        self.contents.save(filename, "PNG")

    def widget(self):
        label = QtGui.QLabel()
        self.contents = self.contents.scaledToHeight(
            self.screen_size, QtCore.Qt.SmoothTransformation)
        label.setPixmap(self.contents)
        return label

    def _draw_line(self, p1, p2, c1=(0, 0, 0), c2=None,
                   op=QtCore.Qt.SolidLine):
        qp1 = QtCore.QPointF(*self._convert(p1))
        qp2 = QtCore.QPointF(*self._convert(p2))
        if c2 and c2 != c1:
            qpm = QtCore.QPointF(
                (qp1.x() + qp2.x()) / 2, (qp1.y() + qp2.y()) / 2)
            pen = QtGui.QPen(QtGui.QColor(*c1), LINE_WEIGHT, op)
            self.painter.setPen(pen)
            self.painter.drawLine(qp1, qpm)
            pen.setColor(QtGui.QColor(*c2))
            self.painter.setPen(pen)
            self.painter.drawLine(qpm, qp2)
        else:
            pen = QtGui.QPen(QtGui.QColor(*c1), LINE_WEIGHT, op)
            self.painter.setPen(pen)
            self.painter.drawLine(qp1, qp2)

    def _draw_dashed_line(self, p1, p2, c1=(0, 0, 0), c2=None):
        self._draw_line(p1, p2, c1, c2, QtCore.Qt.DashLine)

    def _draw_wave_line(self, pos1, pos2, width, color=(0, 0, 0)):
        cpos1 = self._convert(pos1)
        cpos2 = self._convert(pos2)
        b1 = m_seg(cpos1, cpos2, math.pi / -2, width / 2)
        b2 = m_seg(cpos1, cpos2, math.pi / 2, width / 2)
        path = QtGui.QPainterPath()
        path.moveTo(*cpos1)
        for i, b in enumerate([b1, b2, b1, b2, b1, b2]):
            p = scale(b[1], (i + 1) / 7, b[0])
            path.lineTo(*p)
        path.lineTo(*cpos2)
        pen = QtGui.QPen(QtGui.QColor(*color), 1, QtCore.Qt.SolidLine)
        self.painter.setPen(pen)
        self.painter.drawPath(path)

    def _draw_wedge(self, head, tail, width, color=(0, 0, 0)):
        chead = self._convert(head)
        ctail = self._convert(tail)
        b1 = m_seg(ctail, chead, math.pi / -2, width / 2)[0]
        b2 = m_seg(ctail, chead, math.pi / 2, width / 2)[0]
        ps = (chead, b1, b2)
        polygon = QtGui.QPolygonF()
        for ver in ps:
            polygon.append(QtCore.QPointF(*ver))
        pen = QtGui.QPen(QtGui.QColor(*color), 1, QtCore.Qt.SolidLine)
        self.painter.setPen(pen)
        self.painter.setBrush(QtGui.QColor(*color))
        self.painter.drawPolygon(polygon)

    def _draw_dashed_wedge(self, head, tail, width, color=(0, 0, 0)):
        chead = self._convert(head)
        ctail = self._convert(tail)
        b1 = m_seg(ctail, chead, math.pi / -2, width / 2)[0]
        b2 = m_seg(ctail, chead, math.pi / 2, width / 2)[0]
        pen = QtGui.QPen(QtGui.QColor(*color), 1, QtCore.Qt.SolidLine)
        self.painter.setPen(pen)
        step_num = 6
        # TODO: adjust step number
        for i in range(step_num):
            step = distance(chead, ctail) / step_num * i
            trim = 1 / step_num * i
            bp1, bp2 = p_seg(b1, b2, True, step, trim)
            qp1 = QtCore.QPointF(*bp1)
            qp2 = QtCore.QPointF(*bp2)
            self.painter.drawLine(qp1, qp2)

    def _draw_text(self, pos, text, size, color=(0, 0, 0), align="center"):
        qfont = QtGui.QFont("Helvetica", size)
        qtext = QtGui.QTextDocument()
        qtext.setDefaultFont(qfont)
        html_format = "<span style='color:rgb({},{},{})'>{}</span>"
        colored = list(color) + [text, ]
        qtext.setHtml(html_format.format(*colored))
        cpos = self._convert(pos)
        offset = size / 1.5
        if align == "center":
            qpos = QtCore.QPointF(
                cpos[0] - qtext.idealWidth() / 2, cpos[1] - offset)
        elif align == "right":
            qpos = QtCore.QPointF(
                cpos[0] - qtext.idealWidth() + offset, cpos[1] - offset)
        else:
            qpos = QtCore.QPointF(cpos[0] - offset, cpos[1] - offset)
        self.painter.save()
        self.painter.translate(qpos)
        qtext.drawContents(self.painter)
        self.painter.restore()

    def _convert(self, pos):
        """ For QPainter coordinate system, reflect over X axis and
        translate from center to top-left
        """
        px = pos[0] + self.logical_size.width() / 2
        py = self.logical_size.height() / 2 - pos[1]
        return px, py



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

import networkx as nx

from chorus.model.atom import Atom
from chorus.model.bond import Bond


class Compound(object):
    """molecule in graph expression.

    Attributes:
        graph (networkx.Graph): molecule graph
        data (dict): optional attributes(ex. general name, safety info)
        descriptors (set): Available descriptors
    """
    def __init__(self, data=None):
        self.graph = nx.Graph()
        self.data = {}
        self.descriptors = set()

        """ Topology (topology) """
        self.rings = None
        self.scaffolds = None
        self.isolated = None

        """ Canvas size (draw.helper) """
        self.size2d = None

        # loads JSON formatted dict
        if data is not None:
            if "data" in data:
                self.data = data["data"]
            else:  # for backword compatibility (will be removed)
                self.data = data["options"]
            self.descriptors = set(data["descriptors"])
            self.rings = data["rings"]
            self.scaffolds = data["scaffolds"]
            self.isolated = data["isolated"]
            self.size2d = data["size2d"]
            for n, atom in data["atoms"].items():
                self.add_atom(int(n), Atom(atom["sym"], atom))
            for u, conn in data["connections"].items():
                for v, bond in conn.items():
                    self.add_bond(int(u), int(v), Bond(bond))

    def __str__(self):
        """Simple molecule description"""
        adscs = []
        for i, a in self.atoms_iter():
            adsc = "{}({}{} {} {})".format(
                i, a.symbol, a.charge_sign(), int(a.pi), a.stereo)
            adscs.append(adsc)
        atext = ", ".join(adscs)
        odsign = {1: "-", 2: "=", 3: "#"}
        bdscs = []
        for u, v, b in self.bonds_iter():
            bdsc = "{}{}{}".format(u, odsign[b.order], v)
            bdscs.append(bdsc)
        btext = ", ".join(bdscs)
        return "\n".join(["Compound:", atext, btext])

    def __len__(self):
        """Alias of atom_count"""
        return self.atom_count()

    def require(self, desc):
        if desc not in self.descriptors:
            raise TypeError("Descriptor '{}' is required.".format(desc))

    def atom(self, key):
        """Get an atom."""
        return self.graph.nodes[key]["atom"]

    def add_atom(self, key, atom):
        """Set an atom. Existing atom will be overwritten."""
        self.graph.add_node(key, atom=atom)

    def remove_atom(self, key):
        """Remove an atom and adjacent bonds."""
        self.graph.remove_node(key)

    def atoms_iter(self):
        """Iterate over atoms."""
        for n, atom in self.graph.nodes.data("atom"):
            yield n, atom

    def atom_count(self):
        """Get number of atoms."""
        return self.graph.number_of_nodes()

    def bond(self, key1, key2):
        """Get a bond."""
        return self.graph[key1][key2]["bond"]

    def add_bond(self, key1, key2, bond):
        """Set a bond. Existing bond will be overwritten."""
        self.graph.add_edge(key1, key2, bond=bond)

    def remove_bond(self, key1, key2):
        """Remove a bond."""
        self.graph.remove_edge(key1, key2)

    def bonds_iter(self):
        """Iterate over bonds."""
        for u, v, bond in self.graph.edges.data("bond"):
            yield u, v, bond

    def bond_count(self):
        """Return number of bonds."""
        return self.graph.number_of_edges()

    def key_set(self):
        """Get a set of atom keys"""
        return set(self.graph.nodes)

    def neighbors(self, key):
        """Return dict of neighbor atom index and connecting bond."""
        return {n: attr["bond"] for n, attr in self.graph[key].items()}

    def neighbor_count(self, key):
        """Return number of neighbors."""
        return len(self.graph[key])

    def neighbors_iter(self):
        """Iterate over atoms and return its neighbors."""
        for n, adj in self.graph.adj.items():
            yield n, {n: attr["bond"] for n, attr in adj.items()}

    def clear(self):
        """Empty the instance """
        # self.graph = nx.Graph()
        self.graph.clear()
        self.data.clear()
        self.descriptors.clear()
        self.size2d = None
        self.rings = None
        self.scaffolds = None
        self.isolated = None

    def add_molecule(self, mol, bond=None, base=None, target=None):
        """connect atom group (for SMILES parser)

        May requires recalculation of 2D coordinate for drawing

        Args:
            mol: graphmol.Compound()
                the original object will be copied.
            bond: Bond object to be connected.
                the original will not be copied so be careful.
            base: index of atom in self to connect
            target: index of atom in group to be connected
        Raises:
            TypeError
        """
        ai = self.available_idx()
        mapping = {n: n + ai - 1 for n, _ in mol.atoms_iter()}
        relabeled = nx.relabel_nodes(mol.graph, mapping)  # copy=True
        self.graph.add_nodes_from(relabeled.nodes(data=True))
        self.graph.add_edges_from(relabeled.edges(data=True))
        if bond:
            self.add_bond(base, mapping[target], bond)

    def available_idx(self):
        return max(self.graph.nodes, default=0) + 1

    def expand(self, idx):
        """ expand shorthand symbol to graphmol """
        # TODO: not yet implemented
        pass

    def collapse(self, idxs, newidx, symbol):
        pass

    def jsonized(self):
        data = {
            "descriptors": list(self.descriptors),
            "isolated": self.isolated,
            "data": self.data,
            "rings": self.rings,
            "scaffolds": self.scaffolds,
            "size2d": self.size2d
        }
        data["atoms"] = {}
        data["connections"] = {}
        for n, atom in self.atoms_iter():
            data["atoms"][n] = atom.dumps()
        for u, v, bond in self.bonds_iter():
            if u not in data["connections"]:
                data["connections"][u] = {}
            data["connections"][u][v] = bond.dumps()
        return data




from codecs import open
from os import path
from setuptools import setup, find_packages, Extension
import sys

ignore_cython = False
ext = None

try:
    from Cython.Distutils import build_ext
    ext = Extension(
        "chorus.cython.mcsdr", ["chorus/cython/mcsdr.pyx"], language="c++"
    )
except ImportError:
    if ignore_cython:
        from distutils.command.build_ext import build_ext
        ext = Extension(
            "chorus.cython.mcsdr", ["chorus/cython/mcsdr.cpp"]
        )
        print("Cython is not available. Distribute C++ source files.")
    else:
        print("Error: Cython is required.")
        sys.exit()

here = path.abspath(path.dirname(__file__))
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="Chorus",
    version="0.7.7",
    description="Simple chemical structure modeling toolkit",
    long_description=long_description,
    url="https://github.com/mojaie/chorus",
    author="Seiji Matsuoka",
    author_email="mojaie@aol.com",
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6"
    ],
    keywords="molecule-modeling cheminformatics",
    packages=find_packages(exclude=["chorus.test*"]),
    package_data={
        "": ["*.yaml", "resources/test/*.mol", "resources/DrugBank/*.mol"]
    },
    python_requires=">=3.5",
    install_requires=[
        "numpy", "numexpr", "matplotlib", "networkx", "pyyaml"
    ],
    ext_modules=[ext],
    cmdclass={'build_ext': build_ext}
)



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

import os
import yaml

""" Import Wildman-Crippen logP parameters """
with open(os.path.join(os.path.dirname(__file__), "wclogp.yaml")) as file:
    DATA = yaml.load(file.read())


def wclogp(mol):
    return wcparams(mol, "logP")


def wcmr(mol):
    return wcparams(mol, "MR")


def wcparams(mol, type_):
    """
    Calculate Wildman-Crippen logP
    Wildman S., Crippen G., Prediction of Physicochemical Parameters by Atomic
    Contribution, J. Chem. Inf. Model. 39 (1999) 868-873
    """
    try:
        assign_wctype(mol)
    except:
        return "N/A"
    scores = []
    for i, atom in mol.atoms_iter():
        scores.append(float(DATA[type_][atom.wctype]))
        if atom.H_count:
            if atom.symbol == "C":
                scores.append(DATA[type_]["H1"] * atom.H_count)
            elif atom.symbol == "N":
                scores.append(DATA[type_]["H3"] * atom.H_count)
            elif atom.wctype == "O2a":
                scores.append(DATA[type_]["H4"] * atom.H_count)
            else:
                scores.append(DATA[type_]["H2"] * atom.H_count)
    return round(sum(scores), 2)


def assign_wctype(mol):
    """ Assign atom type of Wildman-Crippen logP (Wildman and Crippen 1999) """
    mol.require("Aromatic")
    p_block = ["Al", "B", "Si", "Ga", "Ge", "As", "Se", "Sn", "Te", "Pb",
               "Ne", "Ar", "Kr", "Xe", "Rn"]
    d_block = ["Fe", "Co", "Cu", "Zn", "Tc", "Cd", "Pt", "Au", "Hg", "Gd"]
    for i, atom in mol.atoms_iter():
        nbrs = mol.neighbors(i)
        nbrcnt = len(nbrs)
        nbratms = [mol.atom(i) for i, _ in nbrs.items()]
        isnbrarm = any([mol.atom(i).aromatic for i, _ in nbrs.items()])

        if atom.symbol == "C":
            if atom.aromatic:  # Aromatic carbon
                ex = [(mol.atom(i), b) for i, b in nbrs.items()
                      if not b.aromatic]
                if ex:
                    sbst, sbstb = ex[0]
                    cnosx = {"C": "C21", "N": "C22", "O": "C23", "S": "C24",
                             "F": "C14", "Cl": "C15", "Br": "C16", "I": "C17"}
                    if sbst.symbol not in cnosx:
                        atom.wctype = "C13"
                    elif sbst.aromatic:
                        atom.wctype = "C20"
                    elif sbst.symbol in ("C", "N", "O") and sbstb.order == 2:
                        atom.wctype = "C25"
                    else:
                        atom.wctype = cnosx[sbst.symbol]
                else:
                    if nbrcnt == 2:  # No substituent
                        atom.wctype = "C18"
                    elif nbrcnt == 3:  # Bridgehead
                        atom.wctype = "C19"
            elif atom.pi == 0:  # Aliphatic carbon
                hcnopsx = ["H", "C", "N", "O", "P", "S", "F", "Cl", "Br", "I"]
                if not all([a.symbol in hcnopsx for a in nbratms]):
                    atom.wctype = "C27"
                elif isnbrarm:  # Adjacent to aromatic
                    if atom.H_count == 3:
                        if nbratms[0].symbol == "C":
                            atom.wctype = "C8"
                        else:
                            atom.wctype = "C9"
                    else:
                        numh = {0: "C12", 1: "C11", 2: "C10"}
                        atom.wctype = numh[atom.H_count]
                elif all([a.symbol == "C" for a in nbratms]):
                    if atom.H_count >= 2:
                        atom.wctype = "C1"
                    else:
                        atom.wctype = "C2"
                else:
                    if atom.H_count >= 2:
                        atom.wctype = "C3"
                    else:
                        atom.wctype = "C4"
            elif atom.pi == 1:
                dbi = [i for i, b in nbrs.items() if b.order == 2][0]
                dba = mol.atom(dbi)
                if dba.symbol == "C":
                    if isnbrarm:  # C=C adjacent to aromatic
                        atom.wctype = "C26"
                    else:  # C=C
                        atom.wctype = "C6"
                else:  # C=(!C)
                    atom.wctype = "C5"
                    # Overwrite carbonyl oxygens
                    if dba.symbol == "O":
                        if all(a.symbol != "C" for a in nbratms):
                            mol.atom(dbi).wctype = "O11"
                        elif any(a.aromatic for a in nbratms):
                            mol.atom(dbi).wctype = "O10"
                        else:
                            mol.atom(dbi).wctype = "O9"
            elif atom.pi == 2:
                od = max([b.order for _, b in nbrs.items()])
                if od == 3:  # Alkyne
                    atom.wctype = "C7"
                elif od == 2:  # Allene
                    atom.wctype = "C6"
            if atom.wctype is None:
                atom.wctype = "CS"

        elif atom.symbol == "N":
            if atom.charge == 0:
                if atom.aromatic:
                    atom.wctype = "N11"
                elif atom.pi == 0:  # amine
                    amines = {
                        0: {0: "N7", 1: "N2", 2: "N1"},
                        1: {0: "N8", 1: "N4", 2: "N3"}
                    }
                    atom.wctype = amines[isnbrarm][atom.H_count]
                elif atom.pi == 1:  # imine
                    if atom.H_count:
                        atom.wctype = "N5"
                    else:
                        atom.wctype = "N6"
                elif atom.pi == 2:  # nitrile
                    atom.wctype = "N9"
            elif atom.charge > 0:
                if atom.aromatic:
                    atom.wctype = "N12"
                elif atom.pi == 0:  # ammonium
                    if atom.H_count:
                        atom.wctype = "N10"
                    else:
                        atom.wctype = "N13"
                elif atom.pi == 1:  # iminium
                    atom.wctype = "N13"
                elif atom.pi == 2:
                    if any(a.charge < 1 for a in nbratms):  # 1,3-dipole
                        if all(a.symbol == "N" for a in nbratms):  # azide
                            atom.wctype = "N14"
                        else:  # diazo
                            atom.wctype = "N13"
                    else:  # nitrilium
                        atom.wctype = "N14"
            else:  # nitride
                atom.wctype = "N14"
            if atom.wctype is None:
                atom.wctype = "NS"

        elif atom.symbol == "O":
            if atom.wctype:  # skip if already assigned by carbonyl C
                continue
            if atom.aromatic:
                atom.wctype = "O1"
            elif atom.pi == 0 and atom.charge == 0:
                if atom.H_count == 1:
                    if nbratms[0].symbol in ("O", "S"):  # ROOH, RSOH
                        atom.wctype = "O2a"
                    elif nbratms[0].carbonyl_C:  # Carboxylic acid
                        atom.wctype = "O2a"
                    else:  # Alcohol
                        atom.wctype = "O2"
                elif atom.H_count == 2:  # Water
                    atom.wctype = "O2"
                elif isnbrarm:  # Ether adjacent to aromatic
                    atom.wctype = "O4"
                else:  # Ether
                    atom.wctype = "O3"
            elif nbrcnt == 1:
                nbr = nbratms[0]
                if nbr.aromatic:  # O=Aromatic
                    atom.wctype = "O8"
                elif nbr.symbol in ("N", "O"):  # N-oxide, O2, NO
                    atom.wctype = "O5"
                elif nbr.symbol == "S":  # S-oxide
                    atom.wctype = "O6"
                elif nbr.carbonyl_C and atom.charge < 0:  # Carboxylate anion
                    atom.wctype = "O12"
                else:  # ex. Phosphine oxide
                    atom.wctype = "O7"
            if atom.wctype is None:
                atom.wctype = "OS"

        elif atom.symbol in ("F", "Cl", "Br", "I"):
            if atom.charge:
                atom.wctype = "Hal"
            else:
                atom.wctype = atom.symbol
        elif atom.symbol == "P":
            atom.wctype = "P"
        elif atom.symbol == "S":
            if atom.aromatic:
                atom.wctype = "S3"
            elif atom.charge:
                atom.wctype = "S2"
            else:
                atom.wctype = "S1"
        elif atom.symbol in p_block:
            atom.wctype = "Me1"
        elif atom.symbol in d_block:
            atom.wctype = "Me2"
        elif atom.symbol == "H":  # Explicit hydrogen
            nbr = nbratms[0]
            if nbr.symbol == "C":
                atom.wctype = "H1"
            elif nbr.symbol == "N":
                atom.wctype = "H3"
            else:
                atom.wctype = "H2"
        else:
            raise ValueError("Undefined atom symbol: {}".format(atom.symbol))
    mol.descriptors.add("Wildman-Crippen")



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

from collections import deque

import networkx as nx


def recognize_nx(mol):
    """ NetworkX implementation (for comparison)"""
    mol.rings = nx.cycle_basis(mol.graph)
    mol.isolated = sorted(nx.connected_components(mol.graph),
                          key=len, reverse=True)[1:]
    mol.scaffolds = [x for x in nx.biconnected_components(mol.graph)
                     if len(x) > 2]


def recognize(mol):
    """ Detect cycle basis, biconnected and isolated components (DFS).
    This will add following attribute to the molecule instance object.

    mol.ring: Cycle basis
    mol.scaffold: biconnected components
    mol.isolated: isolated components other than the largest one

    To find minimum set of rings, additionally execute topology.minify_ring.

    Reference:
        networkx cycle_basis function
    """
    g = set(i for i, _ in mol.atoms_iter())
    bccs = {}  # BiConnected Components
    isoc = []  # ISOlated Components
    while g:
        start = g.pop()
        stack = [start]
        pred = {start: start}
        used = {start: set()}
        root = {start: start}
        while stack:
            tail = stack.pop()
            for nbr in mol.neighbors(tail):
                if nbr not in used:  # New node
                    pred[nbr] = tail
                    stack.append(nbr)
                    used[nbr] = {tail}
                    root[nbr] = nbr
                elif nbr in stack:  # Cycle found
                    pn = used[nbr]
                    cyc = [nbr, tail]
                    p = pred[tail]
                    end = pred[nbr]
                    root[nbr] = root[tail] = root[end]
                    while p not in pn:  # Backtrack
                        cyc.append(p)
                        root[p] = root[end]
                        if p in bccs:  # Append scaffold to new cycle
                            if root[end] not in bccs:
                                bccs[root[end]] = []
                            bccs[root[end]].extend(bccs[p])
                            del bccs[p]
                        p = pred[p]
                    cyc.append(p)
                    if root[end] not in bccs:  # Append new cycle to scaffold
                        bccs[root[end]] = []
                    bccs[root[end]].append(cyc)
                    used[nbr].add(tail)
        isoc.append(list(pred.keys()))
        # print(pred)
        g -= set(pred)
    mol.rings = []
    mol.scaffolds = []
    for cycles in bccs.values():
        rcnt = len(mol.rings)
        mol.rings.extend(cycles)
        mol.scaffolds.append(list(range(rcnt, rcnt + len(cycles))))
    mol.isolated = list(sorted(isoc, key=len, reverse=True))[1:]
    mol.descriptors.add("Topology")


def minify_ring(mol, verbose=False):
    """ Minify ring set (similar to SSSR)
    Limitation: this can not correctly recognize minimum rings
    in the case of non-outerplanar graph.
    Note: concept of SSSR is controversial. Roughly reduce the size of
    cycle basis can help some scaffold-based analysis
    """
    mol.require("Topology")
    for cyc_idx in mol.scaffolds:
        rings = deque(sorted([mol.rings[c] for c in cyc_idx], key=len))
        minified = []
        cnt = 0
        while rings:
            # TODO: can be infinite loop ?
            cnt += 1
            if cnt > 100:
                print("Minimization failed")
                print(mol.options)
                mol.descriptors.add("MinifiedRing")
                return
            r = rings.popleft()
            init_r = r
            if verbose:
                print(len(r), "Ring:{}".format(r))
            for m in minified:
                if verbose:
                    print(len(m), "Minified:{}".format(m))
                resolved = resolve_inclusion(r, m)
                if resolved:
                    if verbose:
                        print(len(resolved[0]), len(resolved[1]),
                              "Resolved:{}".format(resolved))
                    r = resolved[0]
            if verbose:
                print(len(r), "New ring:{}\n".format(r))
            if len(r) == len(init_r):  # no longer be able to minified
                minified.append(r)
            else:  # further minification required
                rings.append(r)
        for c in cyc_idx:
            mol.rings[c] = minified.pop()
    mol.descriptors.add("MinifiedRing")


def resolve_inclusion(a, b):
    lt, bg = (a, b)
    if len(lt) > len(bg):
        lt, bg = (bg, lt)
        rev = True
    else:
        rev = False
    isec = set(lt) & set(bg)
    isize = len(isec)
    if isize == 3 and len(lt) == 4:
        pass  # TODO: cubane special case
    elif isize != len(lt) and isize <= len(lt) / 2 + 1:
        return False  # Already minimum
    lq = deque(lt)
    bq = deque(bg)
    while bq[0] not in isec or bq[-1] in isec:
        bq.rotate(1)
    while lq[0] != bq[0]:
        lq.rotate(1)
    if bq[1] != lq[1]:  # Reverse
        lq = deque(reversed(lq))
        while lq[0] != bq[0]:
            lq.rotate(1)
    if list(bq)[:isize] != list(lq)[:isize]:
        return False  # Aboid irregular minification
    bx = list(bq)[isize - 1:]
    lx = list(lq)[isize:]
    new_bg = bx + [bq[0]] + list(reversed(lx))
    if rev:
        lt, new_bg = (new_bg, lt)
    return lt, new_bg



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#


def remove_salt(mol):
    to_be_removed = []
    for n, atom in mol.atoms_iter():
        if not mol.neighbor_count(n) and atom.charge:
            to_be_removed.append(n)
    for n in to_be_removed:
        mol.remove_atom(n)


def remove_water(mol):
    to_be_removed = []
    for n, atom in mol.atoms_iter():
        if not mol.neighbor_count(n) and atom.symbol == "O":
            to_be_removed.append(n)
    for n in to_be_removed:
        mol.remove_atom(n)


def remove_coordinated_metal(mol):
    # TODO: should be renamed to "remove_heavy_metals"
    heavy_metals = ("Fe", "Co")  # TODO: list transition metals
    to_be_removed = []
    for n, atom in mol.atoms_iter():
        if mol.neighbor_count(n) > 4 and atom.symbol in heavy_metals:
            to_be_removed.append(n)
    for n in to_be_removed:
        mol.remove_atom(n)


def remove_sugar(mol):
    pass



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

import itertools


def consecutive(iterable, n):
    """
    consecutive('ABCDEF', 3) --> ABC BCD CDE DEF
    consecutive(itertools.cycle(iter), n) to get looped sequence
    """
    iterators = itertools.tee(iterable, n)
    for i, it in enumerate(iterators):
        for _ in range(i):
            next(it, None)
    return zip(*iterators)


def repeat_each(iterable, n):
    """
    repeat_each('ABCDEF', 3) --> A A A B B B C C C D D D
    """
    for elem in iterable:
        for _ in range(n):
            yield elem


def chunk(iterable, size):
    """
    chunk('ABCDEFG', 3) --> ABC DEF G
    """
    # TODO: only used in gui.mainwindow(deprecated)
    iterator = iter(iterable)
    while size:
        result = []
        try:
            for i in range(size):
                elem = next(iterator)
                result.append(elem)
            yield tuple(result)
        except StopIteration:
            if result:
                yield tuple(result)
            return


def chunk_truncate(iterable, size):
    """
    chunk_truncate('ABCDEFG', 3) --> ABC DEF
    """
    # TODO: only used in kekulize(deprecated)
    args = [iter(iterable)] * size
    return zip(*args)


def chunk_fill(iterable, size, fillvalue=None):
    """
    chunk_fill('ABCDEFG', 3, 'x') --> ABC DEF Gxx
    """
    # TODO: not used
    args = [iter(iterable)] * size
    return itertools.zip_longest(*args, fillvalue=fillvalue)



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

from indigo import Indigo

from chorus.model.graphmol import Compound
from chorus.model.atom import Atom
from chorus.model.bond import Bond
from chorus import molutil

idg = Indigo()
# TODO: stereo, aromatic etc


def to_real_mol(mol):
    """Convert molecule to indigo real molecule"""
    real = idg.createMolecule()
    return indigo_mol(mol, real)


def to_query_mol(mol):
    """Convert molecule to indigo real molecule"""
    q = idg.createQueryMolecule()
    return indigo_mol(mol, q)


def indigo_mol(mol, idgmol):
    key_to_obj = {}
    for k, a in mol.atoms_iter():
        aobj = idgmol.addAtom(a.symbol)
        if a.coords is not None:
            crds = list(a.coords) + [0] * (3 - len(a.coords))
            aobj.setXYZ(*crds)
        key_to_obj[k] = aobj
    for u, v, b in mol.bonds_iter():
        uobj = key_to_obj[u]
        vobj = key_to_obj[v]
        uobj.addBond(vobj, b.order)
    return idgmol


def from_indigo(idgmol, assign_descriptor=True):
    """Convert RDMol to molecule"""
    mol = Compound()
    for atom in idgmol.iterateAtoms():
        key = atom.index()
        a = Atom(atom.symbol())
        a.coords = list(atom.xyz())
        mol.add_atom(key, a)
    for bond in idgmol.iterateBonds():
        u = bond.source()
        v = bond.destination()
        b = Bond()
        b.order = int(bond.bondOrder())
        mol.add_bond(u.index(), v.index(), b)
    if assign_descriptor:
        molutil.assign_descriptors(mol)
    return mol


def fingerprint_similarity(mol1, mol2):
    """Calculate Indigo fingerprint similarity
    """
    idmol1 = to_real_mol(mol1)
    idmol2 = to_real_mol(mol2)
    fp1 = idmol1.fingerprint("sim")
    fp2 = idmol2.fingerprint("sim")
    return round(idg.similarity(fp1, fp2, "tanimoto"), 2)


def fingerprint_dist(mol1, mol2):
    return round(1 - fingerprint_similarity(mol1, mol2), 2)



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

def assign_type(mol, force_recalc=False):
    """ PATTY [Bush et al. J. Inf. Comput. Sci 33 (1993) 756-762]
    TODO: not yet implemented

    1:cation 2:anion 3:donor 4:acceptor
    5:polar 6:hydrophobe 7:others
    """
    if "PATTY" in mol.descriptors and not force_recalc:
        return
    mol.require("Phys_charge")
    for i, atom in mol.atoms_iter():
        # default is 7 (others)
        nbrcnt = mol.neighbor_count(i)
        if atom.charge > 0 or atom.charge_phys > 0 or \
                atom.charge_conj > 0 and not atom.n_oxide:
            atom.type = 1  # cation
        elif atom.charge < 0 or atom.charge_phys < 0 or \
                atom.charge_conj < 0 and not atom.n_oxide:
            atom.type = 2  # anion
        elif atom.symbol == "N":
            if nbrcnt in (1, 2):
                if atom.pi == 2:
                    atom.type = 3  # donor
                elif atom.pi == 1:
                    atom.type = 4  # acceptor
        elif atom.symbol == "O":
            if nbrcnt == 1 and not atom.pi:
                atom.type = 5  # polar
            else:
                atom.type = 4  # acceptor
        elif atom.symbol in ("C", "Si", "S", "Se", "P", "As"):
            ewg = False
            for n, bond in mol.neighbors(i).items():
                natom = mol.atom(n)
                if natom.symbol in ("N", "O", "S") and atom.pi \
                        and not (natom.pi == 2 and mol.neighbor_count(n) == 3):
                    # the sp2 adjacent to neg (but not conj tert amine) is 7
                    ewg = True
                    break
            if not ewg:
                atom.type = 6  # hydrophobes
        elif atom.symbol in ("F", "Cl", "Br", "I") and nbrcnt == 1:
            atom.type = 6  # typical halogens are hydrophobic
    mol.descriptors.add("PATTY")


# TODO:
# Conjugated amine is not h_acceptor ?
# if atom.symbol == "N" and not atom.pi:
#    if any(mol.atom(nbr).pi for nbr in nbrs.keys()):
#        mol.atom(i).h_acceptor = False
# N of triphenylamine is sp2 and thus is not H acceptor
# Diphenylamine is very weak base (pKb=0.79)
# Phenylamine is as strong base as pyridine (pKb=4-5)
# Conjugated O can be sp2? (ex. O in diphenylether is H acceptor?)


# TODO: -O-, -S-, =O, =S (H-acceptor)
# TODO: -OH, -SH (H-donor, H-acceptor)
# TODO: non-conjugated -NH2, -NH- (H-donor, H-acceptor)
# TODO: non-conjugated -N< (H-acceptor)
# TODO: conjugated -NH2, -NH- (H-donor)
# TODO: conjugated -N< (None)
# TODO: primary imine(H-donor, H-acceptor)
# TODO: secondary imine, aromatic N (H-acceptor)
# TODO: halogen(wildcard: skip evaluation)

# TODO: nitro, diazo, nitrile, isocyan, azide, nitrile oxide(center: N)


"""

- assign_pi
double bond -> atom.pi = 1
triple bond -> atom.pi = 2
=C= -> atom.pi = 2

- assign_valence
N, O with hydrogen -> atom.h_donor = True
F, O, N (except sp3 N adj to multiple bond) -> atom.h_acceptor = True

- assign_aromatic (not yet implemented)
double_bond -> +2, pi=0 and lone_pair -> +2
pi=0 -> abort(not aromatic)
4n+2: aromatic=True
TODO: dipole=-1 -> +2

- electron withdrawing group (dipole=-1) (not yet implemented)
>C=O, >C=S, >S=O, >P=O, -NO2, -C#N
ammonium, sulfonate(-M)
p-nitrile, p-pyridyl(-M)
trifluoro, trichrolo(-E?)
electron donating group (dipole=+1)
lone_pair
-I?
amine


- assign_charge (not yet implemented)
assign charge in physiological condition
evaluate atom.charge_phys, atom.h_donor, atom.h_acceptor, atom.pi
non-conjugated amine (N and lone_pair) -> cation
amidine, guanidine (=NH adj to dipole=1) -> cation
oxoacids C(=O)O, P(=O)(O)O (OH adj to dipole=-1) -> anion
thiophenol (S adj to aromatic) -> anion
1,3-dipole (h_acceptor, none, h_acceptor)
X=N=X formula(wrong) -> ok
X=[N+]-[X-] formula -> neutralize charge_phys

# nitro -[N+]([O-])=O  -[N+](=O)[O-]
# azide -[N-]-[N+]#N  -N=[N+]=[N-]
# diazo　>[C-]-[N+]#N  >C=[N+]=[N-]
# nitrile oxide -[C-]=[N+]=O  -C#[N+][O-]
N, S, P, As-oxide
∋N=O(wrong) -> ok
∋[N+][O-] -> neutralize charge_phys
∋P=O -> ok
∋[P+][O-] -> neutralize charge_phys

# N-oxide(quart amine but one nbr is O-)
# P,As-oxide(pi=1 and nbr=4 but pi nbr is O-)

special case
6-membered ring interaction
# 2-(pyridin-2-yl)acetaldehyde(pyrophthalone)
# 1,3-diketone (acetylaceton)
# TODO: sulfoneamide, phosphoneamide (SorP=N)
# TODO: imidazole,  diaminopyrimidine (basic hetero ring)
# TODO: tetrazole, thiazole, muscimol (acidic hetero rings)
"""



#
# (C) 2014-2017 Seiji Matsuoka
# Licensed under the MIT License (MIT)
# http://opensource.org/licenses/MIT
#

from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem import rdFMCS
from rdkit import DataStructs

from chorus.model.graphmol import Compound
from chorus.model.atom import Atom, atom_number
from chorus.model.bond import Bond
from chorus import molutil

# TODO: stereo, aromatic etc


def to_rdmol(mol):
    """Convert molecule to RDMol"""
    rwmol = Chem.RWMol(Chem.MolFromSmiles(''))
    key_to_idx = {}
    bond_type = {1: Chem.BondType.SINGLE,
                 2: Chem.BondType.DOUBLE,
                 3: Chem.BondType.TRIPLE}
    conf = Chem.Conformer(rwmol.GetNumAtoms())
    for k, a in mol.atoms_iter():
        i = rwmol.AddAtom(Chem.Atom(atom_number(a.symbol)))
        key_to_idx[k] = i
        conf.SetAtomPosition(i, a.coords)
    rwmol.AddConformer(conf)
    for u, v, b in mol.bonds_iter():
        ui = key_to_idx[u]
        vi = key_to_idx[v]
        rwmol.AddBond(ui, vi, bond_type[b.order])
    Chem.GetSSSR(rwmol)  # Ring recognition is required for fingerprint
    rwmol.UpdatePropertyCache(strict=False)
    return rwmol.GetMol()


def from_rdmol(rdmol, assign_descriptor=True):
    """Convert RDMol to molecule"""
    mol = Compound()
    conf = rdmol.GetConformer()
    Chem.Kekulize(rdmol)
    for atom in rdmol.GetAtoms():
        key = atom.GetIdx()
        a = Atom(atom.GetSymbol())
        a.coords = conf.GetAtomPosition(key)
        mol.add_atom(key, a)
    for bond in rdmol.GetBonds():
        u = bond.GetBeginAtomIdx()
        v = bond.GetEndAtomIdx()
        b = Bond()
        b.order = int(bond.GetBondTypeAsDouble())
        mol.add_bond(u, v, b)
    if assign_descriptor:
        molutil.assign_descriptors(mol)
    return mol


def morgan_sim(mol1, mol2, radius=2, digit=3):
    """Calculate morgan fingerprint similarity by using RDKit
    radius=2 roughly equivalent to ECFP4
    """
    rdmol1 = to_rdmol(mol1)
    rdmol2 = to_rdmol(mol2)
    fp1 = AllChem.GetMorganFingerprint(rdmol1, radius)
    fp2 = AllChem.GetMorganFingerprint(rdmol2, radius)
    return round(DataStructs.DiceSimilarity(fp1, fp2), digit)


def morgan_dist(mol1, mol2, radius=2, digit=3):
    return round(1 - morgan_sim(mol1, mol2, radius, digit), digit)


def fmcs(mol1, mol2, timeout=2, digit=3):
    if int(timeout) != float(timeout):
        raise TypeError('timeout should be integer')
    rdmol1 = to_rdmol(mol1)
    rdmol2 = to_rdmol(mol2)
    edges1 = rdmol1.GetNumBonds()
    edges2 = rdmol2.GetNumBonds()
    mcs = rdFMCS.FindMCS((rdmol1, rdmol2), timeout=int(timeout))
    # Jaccard-Tanimoto coefficient
    try:
        sim = round(mcs.numBonds / (edges1 + edges2 - mcs.numBonds), digit)
    except ZeroDivisionError:
        sim = 0
    return {
        "mcs_edges": mcs.numBonds,
        "similarity": sim,
        "timeout": mcs.canceled
    }




import sys

from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext
from Cython.Compiler import Options

Options.annotate = True

if sys.platform.startswith("darwin"):  # clang (Mac OS X)
    ext = Extension(
        "mcsdr", sources=["mcsdr.pyx"], language="c++",
        extra_compile_args=["-std=c++11", "-stdlib=libc++"],
        extra_link_args=["-std=c++11", "-stdlib=libc++"]
    )
else:  # gcc
    ext = Extension(
        "mcsdr", sources=["mcsdr.pyx"], language="c++"
    )

setup(cmdclass={'build_ext': build_ext}, ext_modules=[ext])
